/*
 * Segment 5 Intelligente Analyseverfahren, SWE 5.4 Funktionen Fuzzy
 * Copyright (C) 2007-2015 BitCtrl Systems GmbH
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.iav.fuzzylib.fuzzylib;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.NoSuchElementException;

import com.bitctrl.i18n.Messages;
import com.bitctrl.text.CheckText;

/**
 * Fuzzy-Variable bestehend der Menge ihrer Terme inkl. Zugeh&ouml;rigkeiten.
 * Eine Fuzzy-Variable kennt dar&uuml;ber hinaus die linguistische Variable von
 * der sie abstammt.
 *
 * @author BitCtrl Systems GmbH, Falko Schumann
 */
public class FuzzyVariable {

	/**
	 * Linguistische Variable aus der die Fuzzy-Variable hervorgeht.
	 */
	private final LinguistischeVariable linguistischeVariable;

	/**
	 * Name der Fuzzy-Variable f&uuml;r den Fall, dass keine linguistische
	 * Variable angegeben wurde.
	 */
	private String name;

	/**
	 * Die Terme der Fuzzy-Variable.
	 */
	private final List<Term> terme = new ArrayList<Term>();

	/**
	 * Erzeugt eine Fuzzy-Variable ohne linguistische Variable. Diese
	 * Fuzzy-Variable kann nicht defuzzyfiziert werden.
	 *
	 * @param name
	 *            Name der Fuzzy-Variable
	 * @param terme
	 *            Feld von Termbezeichnungen
	 * @throws IllegalArgumentException
	 *             Wenn kein(e) Term(e) angegeben wurden
	 */
	public FuzzyVariable(final String name, final String... terme) {
		if (terme.length == 0) {
			throw new IllegalArgumentException(
					FuzzyLibMsg.LeereTermliste.toString());
		}

		setName(name);
		linguistischeVariable = null;

		for (final String term : terme) {
			this.terme.add(new Term(this, term));
		}
	}

	/**
	 * Konstruiert eine Fuzzy-Variable und merkt sich die dazugeh&ouml;rige
	 * linguistische Variable.
	 *
	 * @param linguistischeVariable
	 *            Linguistische Variable aus der die Fuzzy-Variable hervorgeht
	 */
	FuzzyVariable(final LinguistischeVariable linguistischeVariable) {
		assert linguistischeVariable != null;

		this.linguistischeVariable = linguistischeVariable;
		setName(null);
	}

	/**
	 * {@inheritDoc}
	 *
	 * @see java.lang.Object#clone()
	 */
	@Override
	public FuzzyVariable clone() {
		FuzzyVariable klon;

		if (linguistischeVariable != null) {
			klon = new FuzzyVariable(linguistischeVariable.clone());
		} else {
			klon = new FuzzyVariable(name,
					getTermnamen().toArray(new String[0]));
		}
		klon.name = name;
		for (final Term t : terme) {
			klon.setTerm(t.getName(), t.getZugehoerigkeit());
		}

		return klon;
	}

	/**
	 * Pr&uuml;ft ob ein bestimmter vorhanden in der Fuzzy-Variable enthalten
	 * ist.
	 *
	 * @param term
	 *            Name des zu suchenden Terms
	 * @return {@code true}, wenn der Term gefunden wurde
	 */
	public boolean enthaelt(final String term) {
		for (final Term t : terme) {
			if (t.getName().equals(term)) {
				return true;
			}
		}

		return false;
	}

	/**
	 * Zwei Fuzzy-Variablen sind gleich, wenn sie von der selben linguistischen
	 * Variable abstammen oder beide keine besitzen, den gleichen Namen haben
	 * und die selben Terme in identischer Reihenfolge besitzen.
	 *
	 * {@inheritDoc}
	 *
	 * @see java.lang.Object#equals(Object)
	 */
	@Override
	public boolean equals(final Object obj) {
		if (obj == this) {
			return true;
		} else if (obj instanceof FuzzyVariable) {
			final FuzzyVariable fv = (FuzzyVariable) obj;
			boolean equals = true;

			// Entweder beide ohne oder beide mit linguistischer Variable
			equals &= ((linguistischeVariable == null)
					&& (fv.linguistischeVariable == null))
					|| ((linguistischeVariable != null)
							&& (fv.linguistischeVariable != null));

			// Linguistische Variablen vorhanden? Dann müssen sie gleich sein
			if ((linguistischeVariable != null)
					&& (fv.linguistischeVariable != null)) {
				equals &= getLinguistischeVariable()
						.equals(fv.getLinguistischeVariable());
			}

			// Name der Fuzzy-Variablen müssen gleich sein
			equals &= getName().equals(fv.getName());

			// Terme müssen gleich sein und in gleicher Reihenfolge
			equals &= getTerme().equals(fv.getTerme());

			return equals;
		}

		return false;
	}

	@Override
	public int hashCode() {
		final int hash1 = name != null ? name.hashCode() : 0;
		final int hash2 = linguistischeVariable != null
				? linguistischeVariable.hashCode() : 0;
		return hash1 ^ hash2;
	}

	/**
	 * Gibt die linguistische Variable zur&uuml;ck, zu der die Fuzzy-Variable
	 * geh&ouml;rt.
	 *
	 * @return Linguistische Variable
	 */
	public LinguistischeVariable getLinguistischeVariable() {
		return linguistischeVariable;
	}

	/**
	 * Gibt den Namen der Fuzzy-Variable zur&uuml;ck. Wurde kein Name angegeben,
	 * wird der Name der linguistischen Variable zur&uuml;ckgegeben.
	 *
	 * @return Name der Fuzzy-Variable
	 */
	public String getName() {
		if (name == null) {
			return linguistischeVariable.getName();
		}

		return name;
	}

	/**
	 * Gibt den Term mit dem angegebenen Namen zur&uuml;ck.
	 *
	 * @param term
	 *            Name des Terms
	 * @return Der Term
	 * @throws NoSuchElementException
	 *             Wenn kein Term mit dem Namen existiert
	 */
	public Term getTerm(final String term) {
		for (final Term t : terme) {
			if (t.getName().equals(term)) {
				return t;
			}
		}

		throw new NoSuchElementException(
				Messages.get(FuzzyLibMsg.TermNichtGefunden, name, term));
	}

	/**
	 * Gibt eine Liste mit den Termen der Fuzzy-Variable zur&uuml;ck.
	 *
	 * @return Liste mit Termen
	 */
	public List<Term> getTerme() {
		return new ArrayList<Term>(terme);
	}

	/**
	 * Gibt eine Liste mit den Termnamen der Fuzzy-Variable zur&uuml;ck.
	 *
	 * @return Liste mit Termnamen
	 */
	public List<String> getTermnamen() {
		final List<String> menge = new ArrayList<String>();

		for (final Term t : terme) {
			menge.add(t.getName());
		}

		return menge;
	}

	/**
	 * Gibt die Zugeh&ouml;rigkeit eines Tems zur&uuml;ck.
	 *
	 * @param term
	 *            Name eines Terms
	 * @return Zugeh&ouml;rigkeit
	 * @throws IllegalArgumentException
	 *             Wenn kein Term mit dem Namen existiert
	 */
	public Float getZugehoerigkeit(final String term) {
		final Term t = getTerm(term);

		if (t == null) {
			throw new IllegalArgumentException(
					Messages.get(FuzzyLibMsg.TermNichtGefunden, name, term));
		}

		return t.getZugehoerigkeit();
	}

	/**
	 * Legt den Namen der Fuzzy-Variable fest.
	 *
	 * @param name
	 *            Name oder {@code null}, wenn der Name der linguistischen
	 *            Variable verwendet werden soll
	 * @throws IllegalArgumentException
	 *             Wenn der Name {@code null} ist oder nur aus Leerzeichen
	 *             besteht
	 */
	public void setName(final String name) {
		if (name != null) {
			if (!CheckText.isPrintable(name)
					|| CheckText.containsLineBreak(name)) {
				throw new IllegalArgumentException(
						FuzzyLibMsg.VariablennameUngueltig.toString());
			}

			this.name = name.trim();
		} else if (linguistischeVariable != null) {
			// Rückfallebene: Name der linguistischen Variable
			this.name = null;
		} else {
			throw new IllegalArgumentException(
					FuzzyLibMsg.VariablennameUngueltig.toString());
		}
	}

	/**
	 * Setzt die Zugeh&ouml;rigkeit aller Term der Fuzzy-Variable.
	 *
	 * @param zugehoerigkeit
	 *            Wert f&uuml;r die Zugeh&ouml;rigkeit
	 * @throws IllegalArgumentException
	 *             Wenn kein Term mit dem Namen existiert oder die
	 *             Zugeh&ouml;rigkeit nicht im erlaubten Intervall [0,1] liegt
	 */
	public void setTerm(final Float zugehoerigkeit) {
		if ((zugehoerigkeit != null)
				&& ((zugehoerigkeit < 0) || (zugehoerigkeit > 1))) {
			throw new IllegalArgumentException(Messages
					.get(FuzzyLibMsg.ZugehoerigkeitUngueltig, zugehoerigkeit));
		}

		for (final Term term : terme) {
			term.set(zugehoerigkeit);
		}
	}

	/**
	 * Setzt die Zugeh&ouml;rigkeit eines Terms der Fuzzy-Variable.
	 *
	 * @param termname
	 *            Name des Terms
	 * @param zugehoerigkeit
	 *            Wert f&uuml;r die Zugeh&ouml;rigkeit
	 * @throws NoSuchElementException
	 *             Wenn kein Term mit dem Namen existiert oder die
	 *             Zugeh&ouml;rigkeit nicht im erlaubten Intervall [0,1] liegt
	 */
	public void setTerm(final String termname, final Float zugehoerigkeit) {
		if ((zugehoerigkeit != null)
				&& ((zugehoerigkeit < 0) || (zugehoerigkeit > 1))) {
			throw new IllegalArgumentException(Messages
					.get(FuzzyLibMsg.ZugehoerigkeitUngueltig, zugehoerigkeit));
		}

		final Term term = getTerm(termname);

		term.set(zugehoerigkeit);
	}

	/**
	 * Gibt ein Tupel der Form Name{Term, Term, ...} zur&uuml;ck.
	 *
	 * {@inheritDoc}
	 *
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		assert terme.size() > 0; // Es muss mindestens ein Term existieren

		String variable = getName();

		final Iterator<Term> i = terme.iterator();
		Term term = i.next();
		variable += "{" + term; // erstes Element //$NON-NLS-1$

		while (i.hasNext()) {
			term = i.next();
			variable += ", " + term; //$NON-NLS-1$
		}

		return variable + "}"; //$NON-NLS-1$
	}

	/**
	 * F&uuml;gt den Term hinzu.
	 *
	 * @param term
	 *            Ein neuer Term
	 */
	void addTerm(final Term term) {
		assert!terme.contains(term);

		terme.add(term);
	}

	/**
	 * Prüft ob die Fuzzy-Variable einen gültigen Wert besitzt. Haben alle Terme
	 * die Zugehörigkeit {@code null} ist die Fuzzy-Variable ungültig.
	 *
	 * @return {@code false}, wenn alle Terme die Zugehörigkeit {@code null}
	 *         haben.
	 */
	public boolean isValid() {
		boolean valid;

		valid = false;
		for (final Term t : getTerme()) {
			valid |= t.getZugehoerigkeit() != null;
		}

		return valid;
	}

}
