/*
 * Segment 5 Intelligente Analyseverfahren, SWE 5.4 Funktionen Fuzzy
 * Copyright (C) 2007-2015 BitCtrl Systems GmbH
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.iav.fuzzylib.fuzzylib;

import static com.bitctrl.text.CheckText.containsLineBreak;
import static com.bitctrl.text.CheckText.isPrintable;
import static de.bsvrz.iav.fuzzylib.fuzzylib.Geometrie.flaeche;
import static de.bsvrz.iav.fuzzylib.fuzzylib.Geometrie.ortsvektor;
import static de.bsvrz.iav.fuzzylib.fuzzylib.Geometrie.schnittpunkt;
import static de.bsvrz.iav.fuzzylib.fuzzylib.Geometrie.schwerpunkt;
import static de.bsvrz.iav.fuzzylib.fuzzylib.Geometrie.trapezEcken;
import static de.bsvrz.iav.fuzzylib.fuzzylib.Geometrie.trapezKanten;
import static de.bsvrz.iav.fuzzylib.fuzzylib.Geometrie.trapezStellen;
import static de.bsvrz.iav.fuzzylib.fuzzylib.Geometrie.vektorAddition;
import static de.bsvrz.iav.fuzzylib.fuzzylib.Geometrie.zugehoerigkeit;
import static java.lang.Math.PI;
import static java.lang.Math.acos;
import static java.lang.Math.round;

import java.awt.geom.Line2D;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.Iterator;
import java.util.List;
import java.util.NoSuchElementException;
import java.util.PriorityQueue;
import java.util.Queue;
import java.util.SortedSet;
import java.util.TreeSet;

import com.bitctrl.i18n.Messages;

import de.bsvrz.iav.fuzzylib.fuzzylib.Geometrie.Point2DComparatorX;

/**
 * Linguistische Variable bestehend aus ihrem Namen, Wertebereich und
 * Fuzzy-Sets.
 * <p>
 * Die Angabe des Wertebereichs ist optional. Wird keiner angegeben, wird als
 * Minimum <code>0</code> und als Maximum {@link java.lang.Integer#MAX_VALUE}
 * angenommen. Werden Werte au&szlig;erhalb des Wertebereichs fuzzyfiziert, ist
 * die Zugeh&ouml;rigkeit aller Terme 0.
 * <p>
 * Die Fuzzy-Sets werden nach der zweiten Trapezecke sortiert gesichert. Diese
 * Ordnung wird beim erstmaligen fuzzyfizieren (mit oder ohne Wert) hergestellt.
 *
 * @author BitCtrl Systems GmbH, Falko Schumann
 */
public class LinguistischeVariable {

	/**
	 * Klasse zum Vergleich bzw Sortieren von Fuzzy-Sets.
	 */
	class FuzzySetComparator implements Comparator<FuzzySet> {

		/**
		 * Vergleicht jeweils die zweite Trapezecke der beiden Fuzzy-Sets.
		 *
		 * @param fs1
		 *            ein Fuzzy-Set.
		 * @param fs2
		 *            ein anderes Fuzzy-Set, mit dem verglichen wird.
		 * @return -1, 0 oder 1
		 * @see java.util.Comparator#compare(java.lang.Object, java.lang.Object)
		 */
		@Override
		public int compare(final FuzzySet fs1, final FuzzySet fs2) {
			int i = 0;

			if (fs1.getT2() < fs2.getT2()) {
				i = -1;
			}

			if (fs1.getT2() > fs2.getT2()) {
				i = 1;
			}

			return i;
		}

	}

	/** Sichert die Fuzzy-Sets der linguistischen Variable. */
	private final List<FuzzySet> fuzzySets;

	/** Name der linguistischen Variable. */
	private final String name;

	/** Minimum des Wertebereichs der linguistischen Variable. */
	private int minimum;

	/** Maximum des Wertebereichs der linguistischen Variable. */
	private int maximum;

	/** {@code true}, wenn ein runder Wertebereich angenommen werden soll. */
	private boolean runderWertebereich = false;

	/**
	 * Konstruiert eine linguistische Variable. Es wird intern
	 * {@link LinguistischeVariable#LinguistischeVariable(String, int, int)} mit
	 * minimum = 0 und maximum = {@link Integer#MAX_VALUE} aufgerufen.
	 *
	 * @param name
	 *            Name der linguistischen Variable
	 */
	public LinguistischeVariable(final String name) {
		this(name, 0, Integer.MAX_VALUE);
	}

	/**
	 * Konstruiert eine linguistische Variable. Das Minimum und Maximum, haben
	 * keinen Einfluss auf das Hinzuf&uuml;gen von Fuzzy-Sets. Ein
	 * Wertebereichspr&uuml;fung findet dabei nicht statt. Der Wertebereich wird
	 * einzig bei der Defuzzyfizierung benutzt.
	 *
	 * @param name
	 *            Name der linguistischen Variable
	 * @param minimum
	 *            Minimum des Wertebereichs der Basisvariable f&uuml;r die
	 *            Defuzzyfizierung
	 * @param maximum
	 *            Maximum des Wertebereichs der Basisvariable f&uuml;r die
	 *            Defuzzyfizierung
	 * @throws IllegalArgumentException
	 *             Wenn minimum &gt; maximum oder der Name nicht druckbar ist
	 *             bzw. Zeilenumbr&uuml;che enth&auml;lt
	 */
	public LinguistischeVariable(final String name, final int minimum,
			final int maximum) {
		if (!isPrintable(name) || containsLineBreak(name)) {
			throw new IllegalArgumentException(
					FuzzyLibMsg.VariablennameUngueltig.toString());
		}

		if (minimum > maximum) {
			throw new IllegalArgumentException(Messages
					.get(FuzzyLibMsg.WertebereichUngueltig, minimum, maximum));
		}

		this.name = name.trim();
		this.minimum = minimum;
		this.maximum = maximum;
		fuzzySets = new ArrayList<FuzzySet>();
	}

	/**
	 * F&uuml;gt der linguistischen Variable ein neues Fuzzy-Set hinzu.
	 *
	 * @param set
	 *            Fuzzy-Set
	 * @return <code>false</code>, wenn das Fuzzy-Set bereits existiert und
	 *         deshalb nicht hinzugef&uuml;gt wurde, sonst <code>true</code>
	 */
	public boolean addFuzzySet(final FuzzySet set) {
		if (fuzzySets.contains(set)) {
			return false;
		}

		fuzzySets.add(set);
		return true;
	}

	/**
	 * {@inheritDoc}
	 *
	 * @see java.lang.Object#clone()
	 */
	@Override
	public LinguistischeVariable clone() {
		LinguistischeVariable klon;

		klon = new LinguistischeVariable(name, minimum, maximum);
		for (final FuzzySet fs : fuzzySets) {
			klon.fuzzySets.add(new FuzzySet(fs.getName(), fs.getT1(),
					fs.getT2(), fs.getT3(), fs.getT4()));
		}
		return klon;
	}

	/**
	 * Defuzzyfiziert eine Fuzzy-Variable.
	 *
	 * @param fv
	 *            Ein Fuzzy-Wert als Fuzzy-Variable
	 * @return Defuzzyfizierter Wert oder <code>null</code>, wenn die
	 *         Defuzzyfizierung nicht m&ouml;glich war. Letzteres ist der Fall,
	 *         wenn alle Terme die Zugeh&ouml;rigkeit 0.0 haben.
	 * @throws IllegalArgumentException
	 *             Wenn die Fuzzy-Variable nicht zur linguistischen Variable
	 *             geh&ouml;rt
	 */
	public Integer defuzzyfiziere(final FuzzyVariable fv) {
		if (!equals(fv.getLinguistischeVariable())) {
			throw new IllegalArgumentException(
					FuzzyLibMsg.FalscheFuzzyVariable.toString());
		}

		// Haltepunkte in Queue sammeln und relevante Kanten bestimmen
		final List<Double> stellen = new ArrayList<Double>();
		final List<Line2D> kanten = new ArrayList<Line2D>();
		for (final Term t : fv.getTerme()) {
			if (t.getZugehoerigkeit() > 0) {

				kanten.addAll(Arrays.asList(trapezKanten(t)));
				stellen.addAll(trapezStellen(t));
			}
		}

		// Kantenzug über die Konturen der Fuzzy-Sets bestimmen
		final List<Point2D> kantenzug = scanlineKontur(stellen, kanten);

		if (kantenzug.size() == 0) {
			// Es gibt nichts zu tun
			return null;
		}

		// Schwerpunkt bestimmen
		Point2D s;
		if (isRunderWertebereich()) {
			Queue<Point2D> queue;
			queue = new PriorityQueue<Point2D>(20, new Point2DComparatorX());
			queue.addAll(kantenzug);
			s = new Point2D.Double(defuzzyfiziereSpezialfall(queue), 0);
		} else {
			// Rampen sollen an Wertebereichsgrenze gespiegelt werden
			// Rampen suchen und Kantenzug entsprechend ergänzen
			for (final Term t : fv.getTerme()) {
				if (t.getZugehoerigkeit() > 0) {
					final FuzzySet fs = t.getFuzzySet();
					if (fs.isAbsteigendeRampe() && (fs.getT1() == minimum)) {
						// Absteigende Rampe am Minimum spiegeln
						for (final Point2D p : trapezEcken(t)) {
							if (p.getX() != minimum) {
								kantenzug.add(new Point2D.Double(
										minimum - p.getX(), p.getY()));
							}
						}
					} else if (fs.isAufsteigendeRampe()
							&& (fs.getT4() == maximum)) {
						// Aufsteigende Rampe am Maximum spiegeln
						for (final Point2D p : trapezEcken(t)) {
							if (p.getX() != maximum) {
								kantenzug.add(new Point2D.Double(
										(2 * maximum) - p.getX(), p.getY()));
							}
						}
					}
				}
			}

			// Schwerpunkt der Fläche unter dem Kantenzug (der Kontur) bestimmen
			s = schwerpunkt(new ArrayList<Point2D>(kantenzug));
		}

		if (s == null) {
			return null;
		}
		return round(Double.valueOf(s.getX()).floatValue());
	}

	/**
	 * Pr&uuml;ft ob ein bestimmter vorhanden in der Fuzzy-Variable enthalten
	 * ist.
	 *
	 * @param term
	 *            Name des zu suchenden Terms
	 * @return {@code true}, wenn der Term gefunden wurde
	 */
	public boolean enthaelt(final String term) {
		for (final FuzzySet fs : fuzzySets) {
			if (fs.getName().equals(term)) {
				return true;
			}
		}

		return false;
	}

	/**
	 * Zwei linguistische Variablen sind, wenn sie den gleichen Namen und die
	 * gleiche Menge Fuzzy-Sets besitzen. Die beiden Mengen der Fuzzy-Sets sind
	 * identisch, wenn sie gleich gro&szlig; sind, die gleichen Terme und
	 * Trapezdefinitionen besitzen. Zus&auml;tzlich muss der Wertebereich der
	 * beiden Variablen &uuml;bereinstimmen.
	 *
	 * {@inheritDoc}
	 *
	 * @see java.lang.Object#equals(Object)
	 */
	@Override
	public boolean equals(final Object obj) {
		boolean equals = true;

		if (obj instanceof LinguistischeVariable) {
			final LinguistischeVariable lv = (LinguistischeVariable) obj;

			// Name vergleichen
			if (!name.equals(lv.name)) {
				equals = false;
			}

			// Wertebereich vergleichen
			if ((minimum != lv.minimum) || (maximum != lv.maximum)) {
				equals = false;
			}

			// Menge der Fuzzy-Sets vergleichen
			if (fuzzySets.size() == lv.fuzzySets.size()) {
				for (final FuzzySet fs1 : fuzzySets) {
					if (!lv.enthaelt(fs1.getName())) {
						return false;
					}

					final FuzzySet fs2 = lv.getFuzzySet(fs1.getName());

					if ((fs1.getT1() != fs2.getT1())
							|| (fs1.getT2() != fs2.getT2())
							|| (fs1.getT3() != fs2.getT3())
							|| (fs1.getT4() != fs2.getT4())) {
						equals = false;
					}
				}
			} else {
				equals = false;
			}

		} else {
			equals = false;
		}

		return equals;
	}

	@Override
	public int hashCode() {
		return name.hashCode();
	}

	/**
	 * Erstellt ein Fuzzy-Variable aus der linguistischen Variable. Die
	 * &Uuml;berg&auml;nge zwischen den Fuzzy-Sets werden falls notwendig
	 * berechnet. Die Zugeh&ouml;rigkeiten aller Terme werden mit 0
	 * initialisiert.
	 *
	 * @return Eine Fuzzy-Variable oder <code>null</code>, wenn keine Fuzzy-Sets
	 *         existieren
	 */
	public FuzzyVariable fuzzyfiziere() {
		FuzzyVariable fv = null;

		if (fuzzySets.size() > 0) {
			fv = new FuzzyVariable(this);
			berechneUebergaenge();
			for (final FuzzySet set : fuzzySets) {
				fv.addTerm(new Term(fv, set.getName()));
			}
		}

		return fv;
	}

	/**
	 * Fuzzyfiziert den &uuml;bergebenen Wert. Der Fuzzywert wird mit Hilfe von
	 * {@link LinguistischeVariable#fuzzyfiziere()} erzeugt und
	 * anschlie&szlig;end die Zugeh&ouml;rigkeiten berechnet.
	 *
	 * @param wert
	 *            Zu fuzzyfiziender Wert
	 * @return Der berechnete Fuzzy-Wert als Fuzzy-Variable
	 */
	public FuzzyVariable fuzzyfiziere(final int wert) {
		final FuzzyVariable fv = fuzzyfiziere();

		for (final FuzzySet fs : fuzzySets) {
			if ((minimum <= wert) && (wert <= maximum)) {
				fv.setTerm(fs.getName(), zugehoerigkeit(fs, wert));
			} else {
				fv.setTerm(fs.getName(), 0f);
			}
		}

		return fv;
	}

	/**
	 * Gibt die Liste der Fuzzy-Sets in der linguistischen Variable zur&uuml;ck.
	 *
	 * @return Liste von Fuzzy-Sets
	 */
	public List<FuzzySet> getFuzzySet() {
		return new ArrayList<FuzzySet>(fuzzySets);
	}

	/**
	 * Gibt das Fuzzy-Set mit dem angegebenen Namen zur&uuml;ck.
	 *
	 * @param term
	 *            Name des Fuzzy-Sets
	 * @return Ein Fuzzy-Set oder <code>null</code>, wenn kein Fuzzy-Set mit dem
	 *         Namen existiert
	 * @throws NoSuchElementException
	 *             Wenn kein Fuzzy-Set mit dem Namen existiert
	 */
	public FuzzySet getFuzzySet(final String term) {
		for (final FuzzySet fs : fuzzySets) {
			if (fs.getName().equals(term)) {
				return fs;
			}
		}

		throw new NoSuchElementException(
				Messages.get(FuzzyLibMsg.TermNichtGefunden, name, term));
	}

	/**
	 * Bestimmt falls vorhanden die Definitionsl&uuml;cken.
	 *
	 * @return Punktliste, wobei x den Anfang und y das Ende der L&uuml;cke
	 *         angeben
	 */
	public List<Point2D> getLuecken() {
		if (!hatLuecken()) {
			return null;
		}

		// x: Anfang, y: Ende
		final List<Point2D> luecken = new ArrayList<Point2D>();

		if (fuzzySets.size() == 0) {
			luecken.add(new Point2D.Double(minimum, maximum));
			return luecken;
		}

		// x: Anfang, y: Ende
		final Queue<Point2D> abschnitte = new PriorityQueue<Point2D>(
				fuzzySets.size(), new Point2DComparatorX());

		// Abgedeckte Streckenabschnitte des Wertebereichs sammeln
		for (final FuzzySet fs : fuzzySets) {
			if (fs.isSpezialfall()) {
				abschnitte.add(new Point2D.Float(minimum, fs.getT4()));
				abschnitte.add(new Point2D.Float(fs.getT1(), maximum));
			} else {
				abschnitte.add(new Point2D.Float(fs.getT1(), fs.getT4()));
			}
		}

		Point2D p1, p2;
		p2 = abschnitte.poll();
		while (!abschnitte.isEmpty()) {
			p1 = p2;
			p2 = abschnitte.poll();

			if (p1.getY() < p2.getX()) {
				// Ende liegt vor Anfang des nächsten Streckenabschnitts
				luecken.add(new Point2D.Double(p1.getY(), p2.getX()));
			}
		}

		return luecken;
	}

	/**
	 * Gibt das Maximum des Wertebereichs f&uuml;r die Defuzzyfizierung
	 * zur&uuml;ck.
	 *
	 * @return Maximum
	 */
	public int getMaximum() {
		return maximum;
	}

	/**
	 * Gibt das Minimum des Wertebereichs f&uuml;r die Defuzzyfizierung
	 * zur&uuml;ck.
	 *
	 * @return Minimum
	 */
	public int getMinimum() {
		return minimum;
	}

	/**
	 * Gibt den Namen der linguistischen Variable zur&uuml;ck.
	 *
	 * @return Name der linguistischen Variable
	 */
	public String getName() {
		return name;
	}

	/**
	 * Gibt die Liste der Namen der Fuzzy-Sets in der linguistischen Variable
	 * zur&uuml;ck.
	 *
	 * @return Liste von Termnamen
	 */
	public List<String> getTerme() {
		final List<String> terme = new ArrayList<String>();

		for (final FuzzySet set : fuzzySets) {
			if (!terme.contains(set.getName())) {
				terme.add(set.getName());
			}
		}

		return terme;
	}

	/**
	 * Pr&uuml;ft ob der die Fuzzy-Sets den gesamten Wertebereich
	 * &uuml;berdecken.
	 *
	 * @return {@code false}, wenn es Definitionsl&uuml;cken gibt
	 */
	public boolean hatLuecken() {
		if (fuzzySets.size() == 0) {
			return true;
		}

		// Fuzzy-Sets mit nur zwei Ecken vervollständigen
		berechneUebergaenge();

		// x: Anfang, y: Ende
		final Queue<Point2D> abschnitte = new PriorityQueue<Point2D>(
				fuzzySets.size(), new Point2DComparatorX());

		// Abgedeckte Streckenabschnitte des Wertebereichs sammeln
		for (final FuzzySet fs : fuzzySets) {
			if (fs.isSpezialfall()) {
				abschnitte.add(new Point2D.Float(minimum, fs.getT4()));
				abschnitte.add(new Point2D.Float(fs.getT1(), maximum));
			} else {
				abschnitte.add(new Point2D.Float(fs.getT1(), fs.getT4()));
			}
		}

		Point2D p1, p2;
		p2 = abschnitte.poll();
		if (p2 == null) {
			// Keine Fuzzy-Sets = eine große Lücke
			return true;
		} else if (abschnitte.isEmpty()) {
			// Nur ein Fuzzy-Set ist definiert
			if ((p2.getX() > minimum) || (p2.getY() < maximum)) {
				// Fuzzy-Set berührt/schneidet nicht beide Wertebereichsgrenzen
				return true;
			}
		}

		while (!abschnitte.isEmpty()) {
			p1 = p2;
			p2 = abschnitte.poll();

			if (p1.getY() < p2.getX()) {
				// Ende liegt vor Anfang des nächsten Streckenabschnitts
				return true;
			}
		}

		return false;
	}

	/**
	 * Pr&uuml; ob die linguistische Variable einen runden Wertebereich besitzt.
	 * Dies ist der Fall, wenn es mindestens ein Fuzzy-Set gibt, welches sich
	 * &uuml;ber beide Wertebereichsgrenzen erstreckt. Die Eigenschaft kann auch
	 * explizit gesetzt werden.
	 *
	 * @return {@code true}, wenn ein runder Wertebereich vorliegt
	 * @see de.bsvrz.iav.fuzzylib.fuzzylib.FuzzySet#isSpezialfall()
	 * @see de.bsvrz.iav.fuzzylib.fuzzylib.LinguistischeVariable#setRunderWertebereich(boolean)
	 */
	public boolean isRunderWertebereich() {
		// Festlegung
		if (runderWertebereich) {
			return true;
		}

		// automatische Erkennung
		for (final FuzzySet fs : fuzzySets) {
			if (fs.isSpezialfall()) {
				return true;
			}
		}

		return false;
	}

	/**
	 * Entfernt das Fuzzy-Set mit dem angegebenen Namen. Ist kein Fuzzy-Set mit
	 * dem angegebenen Namen enthalten, passiert nichts.
	 *
	 * @param term
	 *            Name eines Fuzzy-Sets
	 */
	public void removeFuzzySet(final String term) {
		if (!enthaelt(term)) {
			return;
		}

		final FuzzySet set = getFuzzySet(term);

		fuzzySets.remove(set);
	}

	/**
	 * Ersetzt das Fuzzy-Set mit dem angegebenen Namen. Ruft intern nacheinander
	 * {@link LinguistischeVariable#removeFuzzySet(String)} und
	 * {@link LinguistischeVariable#addFuzzySet(FuzzySet)} auf.
	 *
	 * @param term
	 *            Name des zu ersetzenden Fuzzy-Sets
	 * @param set
	 *            Das neue Fuzzy-Set
	 * @return <code>true</code>, wenn das Ersetzen erfolgreich war
	 */
	public boolean replaceFuzzySet(final String term, final FuzzySet set) {
		removeFuzzySet(term);
		return addFuzzySet(set);
	}

	/**
	 * Setzt das Maximum des Wertebereichs f&uuml;r die Defuzzyfizierung.
	 *
	 * @param maximum
	 *            Das neue Maximum
	 */
	public void setMaximum(final int maximum) {
		if (minimum > maximum) {
			throw new IllegalArgumentException(Messages
					.get(FuzzyLibMsg.WertebereichUngueltig, minimum, maximum));
		}

		this.maximum = maximum;
	}

	/**
	 * Setzt das Minimum des Wertebereichs f&uuml;r die Defuzzyfizierung.
	 *
	 * @param minimum
	 *            Das neue Minimum
	 */
	public void setMinimum(final int minimum) {
		if (minimum >= maximum) {
			throw new IllegalArgumentException(Messages
					.get(FuzzyLibMsg.WertebereichUngueltig, minimum, maximum));
		}

		this.minimum = minimum;
	}

	/**
	 * Legt fest, ob der Wertebereich rund interpretiert werden soll. Der
	 * Wertebereich wird automatisch als rund behandelt, wenn ein entsprechender
	 * Spezialfall eines Fuzzy-Sets vorhanden ist. In dem Fall wird ein
	 * {@code false} ignoriert.
	 *
	 * @param runderWertebereich
	 *            Wenn {@code true}, dann fallen Minimum und Maximum aufeinander
	 */
	public void setRunderWertebereich(final boolean runderWertebereich) {
		this.runderWertebereich = runderWertebereich;
	}

	/**
	 * Gibt ein Tupel der Form Variablenname: min-max{FuzzySet, FuzzySet, ...}
	 * zur&uuml;ck.
	 *
	 * {@inheritDoc}
	 *
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		String string = name + "/" + minimum + "-" + maximum; //$NON-NLS-1$ //$NON-NLS-2$

		if (fuzzySets.size() == 0) {
			return string + "(keine Fuzzy-Sets definiert)"; //$NON-NLS-1$
		}

		final Iterator<FuzzySet> i = fuzzySets.iterator();
		FuzzySet set = i.next();
		string += "{" + set; // erstes Element //$NON-NLS-1$

		while (i.hasNext()) {
			set = i.next();
			string += ", " + set; //$NON-NLS-1$
		}

		return string + "}"; //$NON-NLS-1$
	}

	/**
	 * Berechnet falls notwendig, die &Uuml;berg&auml;nge zwischen den
	 * Fuzzy-Sets. F&uuml;r alle Trapeze bei denen nur die beiden inneren Ecken
	 * T2 und T3 angegeben wurden, werden die Ecken T1 und T4 derart berechnet,
	 * dass ein flie&szlig;ender &Uuml;bergang zu den benachbarten Trapezen
	 * entsteht.
	 *
	 * @throws IllegalStateException
	 *             Wenn sich Fuzzy-Sets &uuml;berschneiden und dies nicht
	 *             aufgel&ouml;st werden kann.
	 */
	private void berechneUebergaenge() {
		Collections.sort(fuzzySets, new FuzzySetComparator());
		for (int i = 1; i < fuzzySets.size(); i++) {
			// Ecke t4 von i-1 pr&uuml;fen
			if (fuzzySets.get(i - 1).getT4() == Integer.MAX_VALUE) {
				try {
					fuzzySets.get(i - 1).setT4(fuzzySets.get(i).getT2());
				} catch (final IllegalArgumentException ex) {
					throw new IllegalStateException(Messages.get(
							FuzzyLibMsg.FuzzySetMit2PunktenUngueltig,
							fuzzySets.get(i - 1), fuzzySets.get(i)));
				}
			}

			// Ecke t1 von i pr&uuml;fen
			if (fuzzySets.get(i).getT1() == Integer.MIN_VALUE) {
				try {
					if (fuzzySets.get(i).getT2() <= fuzzySets.get(i).getT3()) {
						fuzzySets.get(i).setT1(fuzzySets.get(i - 1).getT3());
					} else {
						// Spezialfall: Fuzzy-Set &uuml;ber Wertebereichsgrenzen
						fuzzySets.get(0).setT1(fuzzySets.get(i).getT3());
						fuzzySets.get(i).setT1(fuzzySets.get(i - 1).getT3());
						fuzzySets.get(i).setT4(fuzzySets.get(0).getT2());
					}
				} catch (final IllegalArgumentException ex) {
					throw new IllegalStateException(Messages.get(
							FuzzyLibMsg.FuzzySetMit2PunktenUngueltig,
							fuzzySets.get(i - 1), fuzzySets.get(i)));
				}
			}
		}
	}

	/**
	 * Bestimmt den Schwerpunkt der Fl&auml;che unter dem Kantenzug. Es wird
	 * dabei von einem runden Wertebereich ausgegangen.
	 *
	 * @param kantenzug
	 *            Liste der Punkte des Kantenzugs
	 * @return Defuzzyfizierter Wert
	 */
	private Integer defuzzyfiziereSpezialfall(
			final Collection<Point2D> kantenzug) {
		assert kantenzug.size() >= 2;

		Queue<Point2D> queue;
		final List<Point2D> vektoren = new ArrayList<Point2D>();
		Point2D p1, p2, v;
		Double wert;

		queue = new PriorityQueue<Point2D>(kantenzug.size() * 2,
				new Point2DComparatorX());
		queue.addAll(kantenzug);

		// Ortsvektoren der Teilflächen sammeln
		p2 = queue.poll();
		while (!queue.isEmpty()) {
			double a; // Teilfläche
			double sx; // Schwerpunkt der Teilfläche

			p1 = p2;
			p2 = queue.poll();

			if (p1.getY() == p2.getY()) {
				// Unter der Kante p1-p2 ist ein Rechteck
				p1 = new Point2D.Double(p1.getX(), 0); // Auf Grundlinie setzen
				final Rectangle2D r = new Rectangle2D.Double();
				r.setFrameFromDiagonal(p1, p2);
				a = flaeche(r);
				sx = schwerpunkt(r).getX();
				vektoren.add(ortsvektor(sx, a, minimum, maximum));
			} else {
				// Unter der Kante p1-p2 ist ein Trapez
				// Genauer: Ein Dreieck was auf einem Rechteck steht

				Point2D p3; // Dritter Dreieckspunkt
				if (p1.getY() > p2.getY()) {
					p3 = new Point2D.Double(p1.getX(), p2.getY());
				} else {
					p3 = new Point2D.Double(p2.getX(), p1.getY());
				}

				// Dreieck berechnen
				a = flaeche(p1, p2, p3);
				sx = schwerpunkt(p1, p2, p3).getX();
				vektoren.add(ortsvektor(sx, a, minimum, maximum));

				// Rechteck berechnen
				final Rectangle2D r = new Rectangle2D.Double();
				if (p1.getY() > p2.getY()) {
					p3 = new Point2D.Double(p3.getX(), 0); // Auf Grundlinie
					// setzen
					r.setFrameFromDiagonal(p3, p2);
				} else {
					p1 = new Point2D.Double(p1.getX(), 0); // Auf Grundlinie
					// setzen
					r.setFrameFromDiagonal(p1, p3);
				}
				a = flaeche(r);
				sx = schwerpunkt(r).getX();
				vektoren.add(ortsvektor(sx, a, minimum, maximum));
			}
		}

		// Ortvektoren zusammenfassen
		v = vektorAddition(vektoren.toArray(new Point2D[0]));

		if (v.equals(new Point2D.Double())) {
			// Defuzzyfizierung nicht möglich
			return null;
		}

		// Winkel des Vektors bestimmen
		wert = acos(v.getX() / v.distance(0, 0));

		// Winkel auf Wertebereich skalieren
		wert = minimum + ((wert * (maximum - minimum)) / (2 * PI));

		return round(wert.floatValue());
	}

	/**
	 * Benutzt einen Scanline-Algorithmus um die Kontur aller Terme mit
	 * Zugeh&ouml;rigkeit gr&ouml;&szlig;er 0 zu bestimmen. Die
	 * Zugeh&ouml;rigkeit begrenzt die H&ouml;he der einzelnen Fuzzy-Sets.
	 *
	 * @param punkte
	 *            Die Haltepunkte der Scanline
	 * @param kanten
	 *            Die Kanten der geometrischen Figuren
	 * @return Kantenzug &uuml;ber die Kontur der Terme
	 */
	private static List<Point2D> scanlineKontur(final Collection<Double> punkte,
			final List<Line2D> kanten) {
		final Queue<Double> queue = new PriorityQueue<Double>(punkte);
		final SortedSet<Point2D> kantenzug = new TreeSet<Point2D>(
				new Point2DComparatorX());
		while (!queue.isEmpty()) {
			final double d = queue.poll();
			final Line2D scanline = new Line2D.Double(d, 0, d, 1);

			// aktive Kanten finden: alle die scan schneiden oder berühren
			// Ausnahme: senkrechte Geraden werden nicht berücksichtigt
			final List<Line2D> aktiveKanten = new ArrayList<Line2D>();
			for (final Line2D linie : kanten) {
				if ((linie.getX1() <= scanline.getX1())
						&& (scanline.getX1() <= linie.getX2())
						&& (linie.getX1() != linie.getX2())) {
					aktiveKanten.add(linie);
				}
			}

			assert aktiveKanten.size() > 0;

			Point2D neuerPunkt = null;
			for (final Line2D l1 : aktiveKanten) {
				// Nach y-Koordinate höchsten Schnittpunkt mit scan bestimmen
				final Point2D s0 = schnittpunkt(scanline, l1);
				if (neuerPunkt == null) {
					neuerPunkt = s0;
				} else if (s0.getY() > neuerPunkt.getY()) {
					neuerPunkt = s0;
				}

				// Alle Schnittpunkte der Kanten > scan in Queue einfügen
				for (final Line2D l2 : aktiveKanten) {
					if (l1.intersectsLine(l2)) {
						final double p = schnittpunkt(l1, l2).getX();
						if (p > scanline.getX1()) {
							queue.add(p);
						}
					}
				}
			}

			assert neuerPunkt != null;

			// Bei drei aufeinanderfolgende Punkten gleicher Höhe, mittleren
			// entfernen
			if (kantenzug.size() >= 2) {
				final Point2D[] liste = kantenzug.toArray(new Point2D[0]);
				final Point2D vorletzterPunkt = liste[liste.length - 2];
				final Point2D letzterPunkt = liste[liste.length - 1];
				if ((neuerPunkt.getY() == letzterPunkt.getY())
						&& (letzterPunkt.getY() == vorletzterPunkt.getY())) {
					kantenzug.remove(letzterPunkt);
				}
			}

			// Punkt der Kontur hinzufügen
			kantenzug.add(neuerPunkt);
		}
		return new ArrayList<Point2D>(kantenzug);
	}

}
