/*
 * Segment 5 Intelligente Analyseverfahren, SWE 5.4 Funktionen Fuzzy
 * Copyright (C) 2007-2015 BitCtrl Systems GmbH
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.iav.fuzzylib.fuzzylib;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertTrue;

import org.junit.Test;

import de.bsvrz.iav.fuzzylib.VKfz;
import de.bsvrz.iav.fuzzylib.WR;

/**
 * Testet grundlegende Fuzzy-Funktionen, wie das Anlegen von linguistischen und
 * Fuzzy-Variablen sowie Vergleiche.
 *
 * Beim Ausf&uuml;hren des Test ist eine nicht 100%-ige &Uuml;berdeckung des
 * Quelltext normal, da es gew&uuml;nscht ist dass einige Bedingungen (
 * {@code assert}) und Ausnahmen ({@code Exception}) entweder immer auftreten
 * oder nie auftreten.
 *
 * @author BitCtrl Systems GmbH, Falko Schumann
 */
@SuppressWarnings("nls")
public class TestFuzzySet {

	/**
	 * Legt mehrere Fuzzy-Sets an und vergleicht diese.
	 */
	@Test
	@SuppressWarnings("static-method")
	public void fuzzySetEqual() {
		final FuzzySet fsNiedrig = new FuzzySet("niedrig", 0, 0, 25, 35);
		final FuzzySet fsMittel = new FuzzySet("mittel", 25, 35, 75, 85);
		final FuzzySet fsHoch = new FuzzySet("hoch", 75, 85, 254, 254);
		final FuzzySet setA = new FuzzySet("hoch", 0, 0, 0, 0);
		final FuzzySet setB = new FuzzySet("mittel", 0, 0, 0, 0);
		final FuzzySet setC = new FuzzySet("hoch", 0, 0, 0, 0);

		assertFalse(fsNiedrig.equals(setA));
		assertEquals(fsMittel, setB);
		assertEquals(fsHoch, setC);
	}

	/**
	 * Konstruiert Fuzzy-Sets mit verschiedenen Namen.
	 */
	@Test
	@SuppressWarnings("static-method")
	public void fuzzySetKonstruktionNamen() {
		String name;
		final String[] richtig = { "heiß;", "kalt", "warm", "123", "4.56" };
		final String[] falsch = { "", " ", "\n", "\t" };

		assertEquals(richtig[0], new FuzzySet(richtig[0], 0, 0).getName());

		// Zuerst die richtigen, dann die falschen testen
		for (int i = 0; i < (richtig.length + falsch.length); i++) {
			if (i < richtig.length) {
				name = richtig[i];
			} else {
				name = falsch[i - richtig.length];
			}

			try {
				@SuppressWarnings("unused")
				final FuzzySet fs = new FuzzySet(name, 0, 0);
				assertTrue(i < richtig.length);
			} catch (final IllegalArgumentException e) {
				assertTrue(i >= richtig.length);
			}
		}
	}

	/**
	 * Konstruiert Fuzzy-Sets mit verschiedenen Trapezecken.
	 */
	@Test
	@SuppressWarnings("static-method")
	public void fuzzySetKonstruktionTrapezEcken() {
		final String name = "Term";
		int[] ecken;
		final int[][] richtig = { { 0, 10, 20, 30 }, { 0, 10, 10, 20 },
				{ 10, 10, 10, 10 }, { 0, 0, 20, 30 }, { 20, 30, 50, 50 },
				{ /* Spezialfall */40, 50, 10, 20 } };
		final int[][] falsch = { { 50, 40, 30, 20 }, { 40, 50, 20, 10 },
				{ 20, 10, 50, 40 }, { 40, 50, 20, 10 } };

		// Zuerst die richtigen, dann die falschen testen
		for (int i = 0; i < (richtig.length + falsch.length); i++) {
			if (i < richtig.length) {
				ecken = richtig[i];
			} else {
				ecken = falsch[i - richtig.length];
			}

			try {
				@SuppressWarnings("unused")
				final FuzzySet fs = new FuzzySet(name, ecken[0], ecken[1],
						ecken[2], ecken[3]);
				assertTrue(i < richtig.length);
			} catch (final IllegalArgumentException e) {
				assertTrue(i >= richtig.length);
			}
		}
	}

	/**
	 * Modifiziert die Trapezecken eines Fuzzy-Set.
	 */
	@Test
	@SuppressWarnings("static-method")
	public void fuzzySetModifikation() {
		FuzzySet fs;

		// Erste Trapezecke
		fs = new FuzzySet("Term", 10, 20, 30, 40);
		for (int i = 0; i <= 25; i += 5) {
			try {
				fs.setT1(i);
				assertTrue(i <= fs.getT2());
			} catch (final IllegalArgumentException e) {
				assertTrue(i > fs.getT2());
			}
		}

		// Zweite Trapezecke
		fs = new FuzzySet("Term", 10, 20, 30, 40);
		for (int i = 5; i <= 35; i += 5) {
			try {
				fs.setT2(i);

				assertTrue(i >= fs.getT1());
			} catch (final IllegalArgumentException e) {
				assertTrue(i < fs.getT1());
			}
		}

		// Dritte Trapezecke
		fs = new FuzzySet("Term", 10, 20, 30, 40);
		for (int i = 15; i <= 45; i += 5) {
			try {
				fs.setT3(i);

				assertTrue(i <= fs.getT4());
			} catch (final IllegalArgumentException e) {
				assertTrue(i > fs.getT4());
			}
		}

		// Vierte Trapezecke
		fs = new FuzzySet("Term", 10, 20, 30, 40);
		for (int i = 25; i <= 50; i += 5) {
			try {
				fs.setT4(i);
				assertTrue(i >= fs.getT3());
			} catch (final IllegalArgumentException e) {
				assertTrue(i < fs.getT3());
			}
		}
	}

	/**
	 * Testet die isXYZ()-Funktionen des Fuzzy-Sets.
	 */
	@Test
	@SuppressWarnings("static-method")
	public void isFunktionen() {
		FuzzySet fs;

		fs = new FuzzySet("nix", 10, 20, 20, 30);
		assertTrue(fs.isDreieck());
		assertFalse(fs.isTrapez());
		assertFalse(fs.isAufsteigendeRampe());
		assertFalse(fs.isAbsteigendeRampe());
		assertFalse(fs.isSpitze());
		assertFalse(fs.isSpezialfall());

		fs = new FuzzySet("nix", 20, 20, 20, 20);
		assertTrue(fs.isSpitze());
		assertFalse(fs.isTrapez());
		assertFalse(fs.isAufsteigendeRampe());
		assertFalse(fs.isAbsteigendeRampe());
		assertFalse(fs.isDreieck());
		assertFalse(fs.isSpezialfall());

		fs = new VKfz(2).fsMittel;
		assertTrue(fs.isTrapez());
		assertFalse(fs.isDreieck());
		assertFalse(fs.isAufsteigendeRampe());
		assertFalse(fs.isAbsteigendeRampe());
		assertFalse(fs.isSpitze());
		assertFalse(fs.isSpezialfall());

		fs = new VKfz(2).fsNiedrig;
		assertTrue(fs.isAbsteigendeRampe());
		assertFalse(fs.isTrapez());
		assertFalse(fs.isAufsteigendeRampe());
		assertFalse(fs.isDreieck());
		assertFalse(fs.isSpitze());
		assertFalse(fs.isSpezialfall());

		fs = new VKfz(2).fsHoch;
		assertTrue(fs.isAufsteigendeRampe());
		assertFalse(fs.isTrapez());
		assertFalse(fs.isDreieck());
		assertFalse(fs.isAbsteigendeRampe());
		assertFalse(fs.isSpitze());
		assertFalse(fs.isSpezialfall());

		assertTrue(new WR().fsNorden.isSpezialfall());
		assertFalse(new WR().fsNorden.isAufsteigendeRampe());
		assertFalse(new WR().fsNorden.isAbsteigendeRampe());
		assertFalse(new WR().fsNorden.isTrapez());
		assertFalse(new WR().fsNorden.isDreieck());
		assertFalse(new WR().fsNorden.isSpitze());
	}

}
