/*
 * Segment 5 Intelligente Analyseverfahren, SWE 5.4 Funktionen Fuzzy
 * Copyright (C) 2007-2015 BitCtrl Systems GmbH
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.iav.fuzzylib.fuzzylib;

import static de.bsvrz.iav.fuzzylib.fuzzylib.Geometrie.flaeche;
import static de.bsvrz.iav.fuzzylib.fuzzylib.Geometrie.ortsvektor;
import static de.bsvrz.iav.fuzzylib.fuzzylib.Geometrie.schnittpunkt;
import static de.bsvrz.iav.fuzzylib.fuzzylib.Geometrie.schwerpunkt;
import static de.bsvrz.iav.fuzzylib.fuzzylib.Geometrie.strahlensatzAB;
import static de.bsvrz.iav.fuzzylib.fuzzylib.Geometrie.strahlensatzCD;
import static de.bsvrz.iav.fuzzylib.fuzzylib.Geometrie.strahlensatzZA;
import static de.bsvrz.iav.fuzzylib.fuzzylib.Geometrie.strahlensatzZC;
import static de.bsvrz.iav.fuzzylib.fuzzylib.Geometrie.trapezEcken;
import static de.bsvrz.iav.fuzzylib.fuzzylib.Geometrie.trapezKanten;
import static de.bsvrz.iav.fuzzylib.fuzzylib.Geometrie.trapezStellen;
import static de.bsvrz.iav.fuzzylib.fuzzylib.Geometrie.vektorAddition;
import static de.bsvrz.iav.fuzzylib.fuzzylib.Geometrie.zugehoerigkeit;
import static org.junit.Assert.assertEquals;
import static org.junit.Assert.fail;

import java.awt.geom.Line2D;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.util.ArrayList;
import java.util.List;
import java.util.PriorityQueue;
import java.util.Queue;

import org.junit.Test;

import de.bsvrz.iav.fuzzylib.KB;
import de.bsvrz.iav.fuzzylib.VKfz;
import de.bsvrz.iav.fuzzylib.WR;
import de.bsvrz.iav.fuzzylib.fuzzylib.Geometrie.Point2DComparatorY;

/**
 * Testet grundlegende Fuzzy-Funktionen, wie das Anlegen von linguistischen und
 * Fuzzy-Variablen sowie Vergleiche.
 *
 * Beim Ausf&uuml;hren des Test ist eine nicht 100%-ige &Uuml;berdeckung des
 * Quelltext normal, da es gew&uuml;nscht ist dass einige Bedingungen (
 * {@code assert}) und Ausnahmen ({@code Exception}) entweder immer auftreten
 * oder nie auftreten.
 *
 * @author BitCtrl Systems GmbH, Falko Schumann
 */
@SuppressWarnings("nls")
public class TestGeometrie {

	/**
	 * Testet die Bestimmung der Fl&auml;che unter einem Kantenzug.
	 */
	@Test
	@SuppressWarnings("static-method")
	public void flaecheKantenzug() {
		final List<Point2D> kantenzug = new ArrayList<Point2D>();

		kantenzug.add(new Point2D.Double(15, 1));
		kantenzug.add(new Point2D.Double(0, 1));
		kantenzug.add(new Point2D.Double(5, 1));
		kantenzug.add(new Point2D.Double(10, 1));
		kantenzug.add(new Point2D.Double(7, 0));

		assertEquals(12.5, flaeche(kantenzug), 0.00001);
	}

	/**
	 * Testet die Bestimmung der Fl&auml;che unter einem Kantenzug.
	 */
	@Test(expected = IllegalArgumentException.class)
	@SuppressWarnings("static-method")
	public void flaecheKantenzugException() {
		final List<Point2D> kantenzug = new ArrayList<Point2D>();

		kantenzug.add(new Point2D.Double(0, 1));
		flaeche(kantenzug);
	}

	/**
	 * Testet die Sortierung von Punkten nach Y-Koordinaten.
	 */
	@Test
	@SuppressWarnings("static-method")
	public void point2DComparatorY() {
		final Queue<Point2D> liste = new PriorityQueue<Point2D>(10,
				new Point2DComparatorY());
		Point2D p1, p2;

		liste.add(new Point2D.Double(12, 64));
		liste.add(new Point2D.Double(21, 13));
		liste.add(new Point2D.Double(54, 155));
		liste.add(new Point2D.Double(97, 94));
		liste.add(new Point2D.Double(165, 6572));
		liste.add(new Point2D.Double(657, 4));

		p2 = liste.poll();
		while (!liste.isEmpty()) {
			p1 = p2;
			p2 = liste.poll();

			if (p1.getY() > p2.getY()) {
				fail();
			}
		}

	}

	/**
	 * Testet die Bestimmung des Schwerpunkts der Fl&auml;che unter einem
	 * Kantenzug.
	 */
	@Test
	@SuppressWarnings("static-method")
	public void schwerpunktKantenzug() {
		final List<Point2D> kantenzug = new ArrayList<Point2D>();
		Point2D s;

		kantenzug.add(new Point2D.Double(10, 0));
		kantenzug.add(new Point2D.Double(20, 1));
		kantenzug.add(new Point2D.Double(60, 1));
		kantenzug.add(new Point2D.Double(70, 0));

		s = new Point2D.Double(40.0, 7.0 / 15.0);
		assertEquals(Double.valueOf(s.getX()).floatValue(),
				Double.valueOf(schwerpunkt(kantenzug).getX()).floatValue(),
				0.00001);
		assertEquals(Double.valueOf(s.getY()).floatValue(),
				Double.valueOf(schwerpunkt(kantenzug).getY()).floatValue(),
				0.00001);
	}

	/**
	 * Testet die Bestimmung des Schwerpunkts der Fl&auml;che unter einem
	 * Kantenzug.
	 */
	@Test(expected = IllegalArgumentException.class)
	@SuppressWarnings("static-method")
	public void schwerpunktKantenzugException() {
		final List<Point2D> kantenzug = new ArrayList<Point2D>();

		kantenzug.add(new Point2D.Double(0, 1));
		schwerpunkt(kantenzug);
	}

	/**
	 * Testet die Bestimmung des Schwerpunkts der Fl&auml;che unter einem
	 * Kantenzug.
	 */
	@Test
	@SuppressWarnings("static-method")
	public void schwerpunktKantenzugOhneFlaeche() {
		final List<Point2D> kantenzug = new ArrayList<Point2D>();

		kantenzug.add(new Point2D.Double(20, 0));
		kantenzug.add(new Point2D.Double(60, 0));

		assertEquals(null, schwerpunkt(kantenzug));
	}

	/**
	 * Testet die Bestimmung der Trapezecken eines Terms.
	 */
	@Test
	@SuppressWarnings("static-method")
	public void testeTrapezEcken() {
		// Hier reicht ein Test, weil die Funktion trapezKanten benutzt

		final LinguistischeVariable lv = new WR().lv;
		final Term t = lv.fuzzyfiziere(10).getTerm("Norden");

		final List<Point2D> ecken = new ArrayList<Point2D>();
		ecken.add(new Point2D.Double(0, 1));
		ecken.add(new Point2D.Double(20, 1));
		ecken.add(new Point2D.Double(25, 0));
		ecken.add(new Point2D.Double(335, 0));
		ecken.add(new Point2D.Double(340, 1));
		ecken.add(new Point2D.Double(360, 1));

		for (final Point2D p : trapezEcken(t)) {
			if (!ecken.contains(p) || (trapezEcken(t).size() != ecken.size())) {
				fail();
			}
		}
	}

	/**
	 * Testet die Bestimmung der Trapezstellen eines Terms.
	 */
	@Test
	@SuppressWarnings("static-method")
	public void testeTrapezStellen() {
		// Hier reicht ein Test, weil die Funktion trapezKanten benutzt

		final LinguistischeVariable lv = new WR().lv;
		final Term t = lv.fuzzyfiziere(10).getTerm("Norden");

		final List<Double> ecken = new ArrayList<Double>();
		ecken.add(0.0);
		ecken.add(20.0);
		ecken.add(25.0);
		ecken.add(335.0);
		ecken.add(340.0);
		ecken.add(360.0);

		for (final Double p : trapezStellen(t)) {
			if (!ecken.contains(p) || (trapezEcken(t).size() != ecken.size())) {
				fail();
			}
		}
	}

	/**
	 * Testet die Berechnung der Fl&auml;che eines Dreiecks.
	 */
	@Test
	@SuppressWarnings("static-method")
	public void testFlaecheDreieck() {
		Point2D p1, p2, p3;

		p1 = new Point2D.Double(1, 5);
		p2 = new Point2D.Double(1, 5);
		p3 = new Point2D.Double(1, 5);
		assertEquals(0.0, flaeche(p1, p2, p3), 0.00001);

		p1 = new Point2D.Double(1, 1);
		p2 = new Point2D.Double(1, 3);
		p3 = new Point2D.Double(1, 5);
		assertEquals(0.0, flaeche(p1, p2, p3), 0.00001);

		p1 = new Point2D.Double(3, 6);
		p2 = new Point2D.Double(9, 6);
		p3 = new Point2D.Double(4, 10);
		assertEquals(12.0f, Double.valueOf(flaeche(p1, p2, p3)).floatValue(),
				0.00001);

		p1 = new Point2D.Double(3, 6);
		p2 = new Point2D.Double(3, 1);
		p3 = new Point2D.Double(4, 3);
		assertEquals(2.5f, Double.valueOf(flaeche(p1, p2, p3)).floatValue(),
				0.00001);
	}

	/**
	 * Testet die Berechnung der Fl&auml;che eines Rechtecks.
	 */
	@Test
	@SuppressWarnings("static-method")
	public void testFlaecheRechteck() {
		Rectangle2D r;

		r = new Rectangle2D.Double(0, 0, 10, 20);
		assertEquals(200.0, flaeche(r), 0.00001);

		r = new Rectangle2D.Double(0, 0, 10, 0);
		assertEquals(0.0, flaeche(r), 0.00001);

		r = new Rectangle2D.Double(0, 0, 0, 0);
		assertEquals(0.0, flaeche(r), 0.00001);
	}

	/**
	 * Testet die Bestimmung der Fl&auml;che eines Terms.
	 */
	@Test
	@SuppressWarnings("static-method")
	public void testFlaecheTerm() {
		// Fällt kurz aus da lediglich flaeche(kantenzug) aufgerufen wird

		Term t;

		t = new WR().lv.fuzzyfiziere(90).getTerm("Osten");
		assertEquals(45.0, flaeche(t), 0.00001);

		t = new WR().lv.fuzzyfiziere(360).getTerm("Osten");
		assertEquals(0.0, flaeche(t), 0.00001);
	}

	/**
	 * Testet die Bestimmung des Ortsvektor eines Terms.
	 */
	@Test
	@SuppressWarnings("static-method")
	public void testOrtsvektor() {
		// Nur wenige Tests, weil Gleitkommaberechnung zu ungenau und damit
		// assertEquals zu oft fehlschlägt (an 16. Kommastelle 16)

		Point2D v, o;
		double sx, a;
		int min, max;

		min = 0;
		max = 360;

		v = new Point2D.Double(1, 1);
		sx = 45;
		a = Math.sqrt(2);
		o = ortsvektor(sx, a, min, max);
		assertEquals(v.getX(), o.getX(), 0.00001);
		assertEquals(v.getY(), o.getY(), 0.00001);

		v = new Point2D.Double(-1, 1);
		sx = 135;
		a = Math.sqrt(2);
		o = ortsvektor(sx, a, min, max);
		assertEquals(v.getX(), o.getX(), 0.00001);
		assertEquals(v.getY(), o.getY(), 0.00001);
	}

	/**
	 * Testet die Berechnung des Schnittpunkts zweier Strecken.
	 */
	@Test
	@SuppressWarnings("static-method")
	public void testSchnittpunkt() {
		Line2D linie1, linie2;
		Point2D s;

		// Normaler Schnittpunkt
		linie1 = new Line2D.Double(0, 0, 10, 10);
		linie2 = new Line2D.Double(0, 10, 10, 0);
		s = new Point2D.Double(5, 5);
		assertEquals(s, schnittpunkt(linie1, linie2));

		// Eine waagerechte Linie
		linie1 = new Line2D.Double(0, 0, 10, 10);
		linie2 = new Line2D.Double(0, 5, 10, 5);
		s = new Point2D.Double(5, 5);
		assertEquals(s, schnittpunkt(linie1, linie2));
		linie1 = new Line2D.Double(0, 10, 10, 10);
		linie2 = new Line2D.Double(0, 0, 10, 20);
		s = new Point2D.Double(5, 10);
		assertEquals(s, schnittpunkt(linie1, linie2));

		// Eine senkrechte Linie
		linie1 = new Line2D.Double(0, 0, 10, 10);
		linie2 = new Line2D.Double(5, 0, 5, 10);
		s = new Point2D.Double(5, 5);
		assertEquals(s, schnittpunkt(linie1, linie2));
		linie1 = new Line2D.Double(10, 0, 10, 10);
		linie2 = new Line2D.Double(0, 0, 20, 10);
		s = new Point2D.Double(10, 5);
		assertEquals(s, schnittpunkt(linie1, linie2));

		// waagerechte parallele Linien
		linie1 = new Line2D.Double(0, 10, 20, 10);
		linie2 = new Line2D.Double(0, 0, 30, 0);
		assertEquals(null, schnittpunkt(linie1, linie2));

		// senkrechte parallele Linien
		linie1 = new Line2D.Double(0, 0, 0, 10);
		linie2 = new Line2D.Double(10, 0, 0, 30);
		s = null;
		assertEquals(s, schnittpunkt(linie1, linie2));

		// schräge parallele Linien
		linie1 = new Line2D.Double(0, 0, 10, 10);
		linie2 = new Line2D.Double(0, 10, 10, 20);
		assertEquals(null, schnittpunkt(linie1, linie2));

		// Eine Linie entartet als Punkt
		linie1 = new Line2D.Double(0, 0, 0, 0);
		linie2 = new Line2D.Double(0, 0, 10, 20);
		s = new Point2D.Double(0, 0);
		assertEquals(s, schnittpunkt(linie1, linie2));
		linie1 = new Line2D.Double(0, 0, 20, 20);
		linie2 = new Line2D.Double(10, 10, 10, 10);
		s = new Point2D.Double(10, 10);
		assertEquals(s, schnittpunkt(linie1, linie2));

		// Beide Linie entartet als Punkt
		linie1 = new Line2D.Double(10, 20, 10, 20);
		linie2 = new Line2D.Double(10, 20, 10, 20);
		s = new Point2D.Double(10, 20);
		assertEquals(s, schnittpunkt(linie1, linie2));

		// Kein Schnittpunkt
		linie1 = new Line2D.Double(0, 0, 10, 10);
		linie2 = new Line2D.Double(0, 20, 30, 20);
		assertEquals(null, schnittpunkt(linie1, linie2));
	}

	/**
	 * Testet die Berechnung des Schwerpunts f&uuml;r Dreiecke.
	 */
	@Test
	@SuppressWarnings("static-method")
	public void testSchwerpunktDreieck() {
		Point2D p1, p2, p3, s;

		p1 = p2 = p3 = s = new Point2D.Double();
		assertEquals(s, schwerpunkt(p1, p2, p3));

		p1 = new Point2D.Double(18, 60);
		p2 = new Point2D.Double(27, 30);
		p3 = new Point2D.Double(42, 45);
		s = new Point2D.Double(29, 45);
		assertEquals(s, schwerpunkt(p1, p2, p3));
	}

	/**
	 * Testet die Berechnung des Schwerpunts f&uuml;r Rechtecke.
	 */
	@Test
	@SuppressWarnings("static-method")
	public void testSchwerpunktRechteck() {
		Rectangle2D r;
		Point2D s;

		r = new Rectangle2D.Double();
		s = new Point2D.Double();
		assertEquals(s, schwerpunkt(r));

		r = new Rectangle2D.Double(20, 10, 60, 40);
		s = new Point2D.Double(50, 30);
		assertEquals(s, schwerpunkt(r));
	}

	/**
	 * Testet die Bestimmung des Schwerpunkts der Fl&auml;che eines Terms.
	 */
	@Test
	@SuppressWarnings("static-method")
	public void testSchwerpunktTerm() {
		// Fällt kurz aus da lediglich schwerpunkt(kantenzug) aufgerufen wird

		Term t;
		Point2D s;

		t = new VKfz(3).lv.fuzzyfiziere(50).getTerm("Mittel");
		s = new Point2D.Double(57.0, 71.0 / 3.0 / 50.0);
		assertEquals(Double.valueOf(s.getX()).floatValue(),
				Double.valueOf(schwerpunkt(t).getX()).floatValue(), 0.00001);
		assertEquals(Double.valueOf(s.getY()).floatValue(),
				Double.valueOf(schwerpunkt(t).getY()).floatValue(), 0.00001);

		t = new VKfz(3).lv.fuzzyfiziere(200).getTerm("Niedrig");
		assertEquals(null, schwerpunkt(t));
	}

	/**
	 * Testen den Strahlensatz.
	 */
	@Test
	@SuppressWarnings("static-method")
	public void testStrahlensatzAB() {
		Line2D cd, za, zc;

		cd = new Line2D.Double(20, 0, 20, 20);
		za = new Line2D.Double(0, 0, 10, 0);
		zc = new Line2D.Double(0, 0, 20, 0);
		assertEquals(10.0, strahlensatzAB(cd, za, zc), 0.00001);
	}

	/**
	 * Testen den Strahlensatz.
	 */
	@Test(expected = IllegalArgumentException.class)
	@SuppressWarnings("static-method")
	public void testStrahlensatzABException() {
		strahlensatzAB(new Line2D.Double(), new Line2D.Double(),
				new Line2D.Double());
	}

	/**
	 * Testen den Strahlensatz.
	 */
	@Test
	@SuppressWarnings("static-method")
	public void testStrahlensatzCD() {
		Line2D ab, za, zc;

		ab = new Line2D.Double(10, 0, 10, 10);
		za = new Line2D.Double(0, 0, 10, 0);
		zc = new Line2D.Double(0, 0, 20, 0);
		assertEquals(20.0, strahlensatzCD(ab, za, zc), 0.00001);
	}

	/**
	 * Testen den Strahlensatz.
	 */
	@Test(expected = IllegalArgumentException.class)
	@SuppressWarnings("static-method")
	public void testStrahlensatzCDException() {
		strahlensatzCD(new Line2D.Double(), new Line2D.Double(),
				new Line2D.Double());
	}

	/**
	 * Testen den Strahlensatz.
	 */
	@Test
	@SuppressWarnings("static-method")
	public void testStrahlensatzZA() {
		Line2D ab, cd, zc;

		ab = new Line2D.Double(10, 0, 10, 10);
		cd = new Line2D.Double(20, 0, 20, 20);
		zc = new Line2D.Double(0, 0, 20, 0);
		assertEquals(10.0, strahlensatzZA(ab, cd, zc), 0.00001);
	}

	/**
	 * Testen den Strahlensatz.
	 */
	@Test(expected = IllegalArgumentException.class)
	@SuppressWarnings("static-method")
	public void testStrahlensatzZAException() {
		strahlensatzZA(new Line2D.Double(), new Line2D.Double(),
				new Line2D.Double());
	}

	/**
	 * Testen den Strahlensatz.
	 */
	@Test
	@SuppressWarnings("static-method")
	public void testStrahlensatzZC() {
		Line2D ab, cd, za;

		ab = new Line2D.Double(10, 0, 10, 10);
		cd = new Line2D.Double(20, 0, 20, 20);
		za = new Line2D.Double(0, 0, 10, 0);
		assertEquals(20.0, strahlensatzZC(ab, cd, za), 0.00001);
	}

	/**
	 * Testen den Strahlensatz.
	 */
	@Test(expected = IllegalArgumentException.class)
	@SuppressWarnings("static-method")
	public void testStrahlensatzZCException() {
		strahlensatzZC(new Line2D.Double(), new Line2D.Double(),
				new Line2D.Double());
	}

	/**
	 * Testet die Bestimmung der Kanten eines Terms, was eine absteigende Rampe
	 * darstellt.
	 */
	@Test
	@SuppressWarnings("static-method")
	public void testTrapezKantenAbsteigendeRampe() {
		final Term t = new KB(6).lv.fuzzyfiziere(15).getTerm("Niedrig");

		final List<Line2D> kanten = new ArrayList<Line2D>();
		kanten.add(new Line2D.Double(0, 0, 0, 1));
		kanten.add(new Line2D.Double(0, 1, 35, 1));
		kanten.add(new Line2D.Double(35, 1, 45, 0));

		for (final Line2D l1 : trapezKanten(t)) {
			boolean gefunden = false;

			for (final Line2D l2 : kanten) {
				if (l1.getP1().equals(l2.getP1())
						&& l1.getP2().equals(l2.getP2())) {
					gefunden = true;
				}
			}

			if (!gefunden || (trapezKanten(t).length != kanten.size())) {
				fail();
			}
		}

	}

	/**
	 * Testet die Bestimmung der Kanten eines Terms, was ein aufsteigende Rampe
	 * darstellt.
	 */
	@Test
	@SuppressWarnings("static-method")
	public void testTrapezKantenAufsteigendeRampe() {
		final Term t = new KB(6).lv.fuzzyfiziere(100).getTerm("Sehr hoch");

		final List<Line2D> kanten = new ArrayList<Line2D>();
		kanten.add(new Line2D.Double(60, 0, 80, 1));
		kanten.add(new Line2D.Double(80, 1, 10000, 1));
		kanten.add(new Line2D.Double(10000, 1, 10000, 0));

		for (final Line2D l1 : trapezKanten(t)) {
			boolean gefunden = false;

			for (final Line2D l2 : kanten) {
				if (l1.getP1().equals(l2.getP1())
						&& l1.getP2().equals(l2.getP2())) {
					gefunden = true;
				}
			}

			if (!gefunden || (trapezKanten(t).length != kanten.size())) {
				fail();
			}
		}

	}

	/**
	 * Testet die Bestimmung der Kanten eines Terms, was ein Dreieck darstellt.
	 */
	@Test
	@SuppressWarnings("static-method")
	public void testTrapezKantenDreieck() {
		final LinguistischeVariable lv = new LinguistischeVariable("egal");
		lv.addFuzzySet(new FuzzySet("unwichtig", 10, 20, 20, 30));
		final Term t = lv.fuzzyfiziere(20).getTerm("unwichtig");

		final List<Line2D> kanten = new ArrayList<Line2D>();
		kanten.add(new Line2D.Double(10, 0, 20, 1));
		kanten.add(new Line2D.Double(20, 1, 20, 1));
		kanten.add(new Line2D.Double(20, 1, 30, 0));

		for (final Line2D l1 : trapezKanten(t)) {
			boolean gefunden = false;

			for (final Line2D l2 : kanten) {
				if (l1.getP1().equals(l2.getP1())
						&& l1.getP2().equals(l2.getP2())) {
					gefunden = true;
				}
			}

			if (!gefunden || (trapezKanten(t).length != kanten.size())) {
				fail();
			}
		}

	}

	/**
	 * Testet die Bestimmung der Kanten eines Terms, der ein Speziallfall
	 * darstellt.
	 */
	@Test
	@SuppressWarnings("static-method")
	public void testTrapezKantenSpezialfall() {
		final LinguistischeVariable lv = new WR().lv;
		final Term t = lv.fuzzyfiziere(10).getTerm("Norden");

		final List<Line2D> kanten = new ArrayList<Line2D>();
		kanten.add(new Line2D.Double(0, 1, 20, 1));
		kanten.add(new Line2D.Double(20, 1, 25, 0));
		kanten.add(new Line2D.Double(335, 0, 340, 1));
		kanten.add(new Line2D.Double(340, 1, 360, 1));

		for (final Line2D l1 : trapezKanten(t)) {
			boolean gefunden = false;

			for (final Line2D l2 : kanten) {
				if (l1.getP1().equals(l2.getP1())
						&& l1.getP2().equals(l2.getP2())) {
					gefunden = true;
				}
			}

			if (!gefunden || (trapezKanten(t).length != kanten.size())) {
				fail();
			}
		}
	}

	/**
	 * Testet die Bestimmung der Kanten eines Terms, was eine Spitze darstellt.
	 */
	@Test
	@SuppressWarnings("static-method")
	public void testTrapezKantenSpitze() {
		final LinguistischeVariable lv = new LinguistischeVariable("egal");
		lv.addFuzzySet(new FuzzySet("unwichtig", 10, 10, 10, 10));
		final Term t = lv.fuzzyfiziere(10).getTerm("unwichtig");

		final List<Line2D> kanten = new ArrayList<Line2D>();
		kanten.add(new Line2D.Double(10, 0, 10, 1));
		kanten.add(new Line2D.Double(10, 1, 10, 1));
		kanten.add(new Line2D.Double(10, 1, 10, 0));

		for (final Line2D l1 : trapezKanten(t)) {
			boolean gefunden = false;

			for (final Line2D l2 : kanten) {
				if (l1.getP1().equals(l2.getP1())
						&& l1.getP2().equals(l2.getP2())) {
					gefunden = true;
				}
			}

			if (!gefunden || (trapezKanten(t).length != kanten.size())) {
				fail();
			}
		}

	}

	/**
	 * Testet die Bestimmung der Kanten eines Terms, was ein Trapez darstellt.
	 */
	@Test
	@SuppressWarnings("static-method")
	public void testTrapezKantenTrapez() {
		final Term t = new KB(6).lv.fuzzyfiziere(50).getTerm("Hoch");

		final List<Line2D> kanten = new ArrayList<Line2D>();
		kanten.add(new Line2D.Double(35, 0, 45, 1));
		kanten.add(new Line2D.Double(45, 1, 60, 1));
		kanten.add(new Line2D.Double(60, 1, 80, 0));

		for (final Line2D l1 : trapezKanten(t)) {
			boolean gefunden = false;

			for (final Line2D l2 : kanten) {
				if (l1.getP1().equals(l2.getP1())
						&& l1.getP2().equals(l2.getP2())) {
					gefunden = true;
				}
			}

			if (!gefunden || (trapezKanten(t).length != kanten.size())) {
				fail();
			}
		}

	}

	/**
	 * Testet die Bestimmung der Trapezkanten eines Terms.
	 */
	@Test(expected = IllegalArgumentException.class)
	@SuppressWarnings("static-method")
	public void testTrapezKantenTrapezException() {
		final Term t = new KB(6).lv.fuzzyfiziere(0).getTerm("Hoch");

		trapezKanten(t);
	}

	/**
	 * Testet die Vektoraddition.
	 */
	@Test
	@SuppressWarnings("static-method")
	public void testVektorAddition() {
		final List<Point2D> punkte = new ArrayList<Point2D>();
		Point2D p;

		punkte.add(new Point2D.Double(2, 0));
		punkte.add(new Point2D.Double(0, -2));
		punkte.add(new Point2D.Double(0, 4));
		punkte.add(new Point2D.Double(-3, -2));
		p = new Point2D.Double(-1, 0);
		assertEquals(p, vektorAddition(punkte.toArray(new Point2D[0])));
	}

	/**
	 * Testet die Berechnung der Zugeh&ouml;rigkeit f&uuml;r eine absteigende
	 * Rampe.
	 */
	@Test
	@SuppressWarnings("static-method")
	public void testZugehoerigkeitAbsteigendeRampe() {
		final FuzzySet fs = new VKfz(6).fsNiedrig;

		// Ecken
		assertEquals(1.0f, zugehoerigkeit(fs, 0), 0.00001);
		assertEquals(1.0f, zugehoerigkeit(fs, 0), 0.00001);
		assertEquals(1.0f, zugehoerigkeit(fs, 23), 0.00001);
		assertEquals(0.0f, zugehoerigkeit(fs, 33), 0.00001);

		// Aufsteigende Kante
		assertEquals(1.0f, zugehoerigkeit(fs, 0), 0.00001);

		// Horizontale Kante
		assertEquals(1.0f, zugehoerigkeit(fs, 20), 0.00001);

		// Absteigende Kante
		assertEquals(0.8f, zugehoerigkeit(fs, 25), 0.00001);

		// Davor
		assertEquals(0.0f, zugehoerigkeit(fs, -6), 0.00001);

		// Dahinter
		assertEquals(0.0f, zugehoerigkeit(fs, 34), 0.00001);
	}

	/**
	 * Testet die Berechnung der Zugeh&ouml;rigkeit f&uuml;r eine aufsteigende
	 * Rampe.
	 */
	@Test
	@SuppressWarnings("static-method")
	public void testZugehoerigkeitAufsteigendeRampe() {
		final FuzzySet fs = new VKfz(6).fsHoch;

		// Ecken
		assertEquals(0.0f, zugehoerigkeit(fs, 73), 0.00001);
		assertEquals(1.0f, zugehoerigkeit(fs, 83), 0.00001);
		assertEquals(1.0f, zugehoerigkeit(fs, 254), 0.00001);
		assertEquals(1.0f, zugehoerigkeit(fs, 254), 0.00001);

		// Aufsteigende Kante
		assertEquals(0.1f, zugehoerigkeit(fs, 74), 0.00001);

		// Horizontale Kante
		assertEquals(1.0f, zugehoerigkeit(fs, 200), 0.00001);

		// Absteigende Kante
		assertEquals(1.0f, zugehoerigkeit(fs, 254), 0.00001);

		// Davor
		assertEquals(0.0f, zugehoerigkeit(fs, 20), 0.00001);

		// Dahinter
		assertEquals(0.0f, zugehoerigkeit(fs, 255), 0.00001);
	}

	/**
	 * Testet die Berechnung der Zugeh&ouml;rigkeit f&uuml;r ein Dreieck.
	 */
	@Test
	@SuppressWarnings("static-method")
	public void testZugehoerigkeitDreieck() {
		final FuzzySet fs = new FuzzySet("unwichtig", 20, 30, 30, 40);

		// Ecken
		assertEquals(0.0f, zugehoerigkeit(fs, 20), 0.00001);
		assertEquals(1.0f, zugehoerigkeit(fs, 30), 0.00001);
		assertEquals(1.0f, zugehoerigkeit(fs, 30), 0.00001);
		assertEquals(0.0f, zugehoerigkeit(fs, 40), 0.00001);

		// Aufsteigende Kante
		assertEquals(0.3f, zugehoerigkeit(fs, 23), 0.00001);

		// Horizontale Kante
		assertEquals(1.0f, zugehoerigkeit(fs, 30), 0.00001);

		// Absteigende Kante
		assertEquals(0.8f, zugehoerigkeit(fs, 32), 0.00001);

		// Davor
		assertEquals(0.0f, zugehoerigkeit(fs, 19), 0.00001);

		// Dahinter
		assertEquals(0.0f, zugehoerigkeit(fs, 255), 0.00001);
	}

	/**
	 * Testet die Berechnung der Zugeh&ouml;rigkeit f&uuml;r den Spezialfall.
	 */
	@Test
	@SuppressWarnings("static-method")
	public void testZugehoerigkeitSpezialfall() {
		final FuzzySet fs = new FuzzySet("unwichtig", 80, 100, 30, 50);

		// Ecken
		assertEquals(0.0f, zugehoerigkeit(fs, 80), 0.00001);
		assertEquals(1.0f, zugehoerigkeit(fs, 100), 0.00001);
		assertEquals(1.0f, zugehoerigkeit(fs, 30), 0.00001);
		assertEquals(0.0f, zugehoerigkeit(fs, 50), 0.00001);

		// Aufsteigende Kante
		assertEquals(0.5f, zugehoerigkeit(fs, 90), 0.00001);

		// Horizontale Kante
		assertEquals(1.0f, zugehoerigkeit(fs, 20), 0.00001);
		assertEquals(1.0f, zugehoerigkeit(fs, 150), 0.00001);

		// Absteigende Kante
		assertEquals(0.85f, zugehoerigkeit(fs, 33), 0.00001);

		// Davor
		assertEquals(0.0f, zugehoerigkeit(fs, 55), 0.00001);

		// Dahinter
		assertEquals(0.0f, zugehoerigkeit(fs, 70), 0.00001);
	}

	/**
	 * Testet die Berechnung der Zugeh&ouml;rigkeit f&uuml;r eine Spitze.
	 */
	@Test
	@SuppressWarnings("static-method")
	public void testZugehoerigkeitSpitze() {
		final FuzzySet fs = new FuzzySet("unwichtig", 50, 50, 50, 50);

		// Ecken
		assertEquals(1.0f, zugehoerigkeit(fs, 50), 0.00001);
		assertEquals(1.0f, zugehoerigkeit(fs, 50), 0.00001);
		assertEquals(1.0f, zugehoerigkeit(fs, 50), 0.00001);
		assertEquals(1.0f, zugehoerigkeit(fs, 50), 0.00001);

		// Aufsteigende Kante
		assertEquals(1.0f, zugehoerigkeit(fs, 50), 0.00001);

		// Horizontale Kante
		assertEquals(1.0f, zugehoerigkeit(fs, 50), 0.00001);

		// Absteigende Kante
		assertEquals(1.0f, zugehoerigkeit(fs, 50), 0.00001);

		// Davor
		assertEquals(0.0f, zugehoerigkeit(fs, 19), 0.00001);

		// Dahinter
		assertEquals(0.0f, zugehoerigkeit(fs, 255), 0.00001);
	}

	/**
	 * Testet die Berechnung der Zugeh&ouml;rigkeit f&uuml;r ein Trapez.
	 */
	@Test
	@SuppressWarnings("static-method")
	public void testZugehoerigkeitTrapez() {
		final FuzzySet fs = new VKfz(6).fsMittel;

		// Ecken
		assertEquals(0.0f, zugehoerigkeit(fs, 23), 0.00001);
		assertEquals(1.0f, zugehoerigkeit(fs, 33), 0.00001);
		assertEquals(1.0f, zugehoerigkeit(fs, 73), 0.00001);
		assertEquals(0.0f, zugehoerigkeit(fs, 83), 0.00001);

		// Aufsteigende Kante
		assertEquals(0.7f, zugehoerigkeit(fs, 30), 0.00001);

		// Horizontale Kante
		assertEquals(1.0f, zugehoerigkeit(fs, 50), 0.00001);

		// Absteigende Kante
		assertEquals(0.5f, zugehoerigkeit(fs, 78), 0.00001);

		// Davor
		assertEquals(0.0f, zugehoerigkeit(fs, 20), 0.00001);

		// Dahinter
		assertEquals(0.0f, zugehoerigkeit(fs, 90), 0.00001);
	}

}
