/*
 * Segment 5 Intelligente Analyseverfahren, SWE 5.4 Funktionen Fuzzy
 * Copyright (C) 2007-2015 BitCtrl Systems GmbH
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.iav.fuzzylib.fuzzylib.inferenz;

import static de.bsvrz.iav.fuzzylib.interpreter.logik.FuzzyLogikHandler.MINIMUM;

import java.util.ArrayList;
import java.util.List;
import java.util.Set;

import de.bsvrz.iav.fuzzylib.fuzzylib.FuzzyVariable;
import de.bsvrz.sys.funclib.bitctrl.interpreter.Ausdruck;
import de.bsvrz.sys.funclib.bitctrl.interpreter.Kontext;
import de.bsvrz.sys.funclib.bitctrl.interpreter.OperationsSymbol;
import de.bsvrz.sys.funclib.bitctrl.interpreter.Operator;
import de.bsvrz.sys.funclib.bitctrl.interpreter.logik.LogischerWert;
import de.bsvrz.sys.funclib.bitctrl.interpreter.logik.LogischesLiteral;

/**
 * Repr&auml;sentiert eine Regel der Fuzzy-Wissensbasis.
 *
 * @author BitCtrl Systems GmbH, Falko Schumann
 */
public class Regel implements Ausdruck {

	/** Die Regelpr&auml;misse. */
	private final Ausdruck praemisse;

	/** Die Konklusion der Regel. */
	private final TermSymbol konklusion;

	/** Der Sicherheitsgrad der Regel. */
	private final Float sicherheitsgrad;

	/** Operator f&uuml;r die Implikation (Standard: Minimum). */
	private Operator opImplikation = MINIMUM;

	/**
	 * Analog {@code Regel(praemisse, new TermSymbol(var, term)}.
	 *
	 * @param praemisse
	 *            Regelpr&auml;misse
	 * @param konklussionVariable
	 *            Fuzzy-Variable der Konklusion
	 * @param konklusionTerm
	 *            Term der Konklusion
	 */
	public Regel(final Ausdruck praemisse, final String konklussionVariable,
			final String konklusionTerm) {
		this(praemisse, new TermSymbol(konklussionVariable, konklusionTerm));
	}

	/**
	 * Wie {@code Regel(praemisse, var, term)}, aber mit Sicherheitsgrad.
	 *
	 * @param praemisse
	 *            Regelpr&auml;misse
	 * @param konklussionVariable
	 *            Fuzzy-Variable der Konklusion
	 * @param konklusionTerm
	 *            Term der Konklusion
	 * @param sicherheitsgrad
	 *            Der Sicheheitsgrad der Regel
	 */
	public Regel(final Ausdruck praemisse, final String konklussionVariable,
			final String konklusionTerm, final Float sicherheitsgrad) {
		this(praemisse, new TermSymbol(konklussionVariable, konklusionTerm),
				sicherheitsgrad);
	}

	/**
	 * Konstruiert eine Regel. Als Sicherheitsgrad wird 1,0 angenommen.
	 *
	 * @param praemisse
	 *            Pr&auml;misse
	 * @param konklusion
	 *            Konklusion
	 */
	public Regel(final Ausdruck praemisse, final TermSymbol konklusion) {
		this(praemisse, konklusion, 1.0f);
	}

	/**
	 * Konstruiert eine Regel.
	 *
	 * @param praemisse
	 *            Pr&auml;misse
	 * @param konklusion
	 *            Konklusion
	 * @param sicherheitsgrad
	 *            Sicherheitsgrad
	 */
	public Regel(final Ausdruck praemisse, final TermSymbol konklusion,
			final Float sicherheitsgrad) {
		assert praemisse != null : "Die Prämisse einer Regel darf nicht null sein."; //$NON-NLS-1$
		assert konklusion != null : "Die Konklusion einer Regel darf nicht null sein."; //$NON-NLS-1$

		this.praemisse = praemisse;
		this.konklusion = konklusion;
		this.sicherheitsgrad = sicherheitsgrad;
	}

	/**
	 * Gibt die Variable der Regelkonklusion zur&uuml;ck.
	 *
	 * @return Der Variablenname
	 */
	public String getKonklusion() {
		return konklusion.getName();
	}

	/**
	 * Gibt die Regel als Ausdruck zur&uuml;ck. Dabei werden Pr&auml;misse und
	 * Konklusion mit dem Implikationsoperator verkn&uuml;pft. Die
	 * zur&uuml;ckgegebene Liste enthält nur einen Ausdruck. Der Ausdruck
	 * besteht aus zwei Elementen bzw. Nachfolgern, der Pr&auml;misse als erstes
	 * Element und der Konklusion als zweites.
	 * <p>
	 * {@inheritDoc}
	 */
	@Override
	public List<Ausdruck> getNachfolger() {
		List<Ausdruck> liste;
		Ausdruck regel;

		regel = new OperationsSymbol(opImplikation, praemisse, konklusion);
		liste = new ArrayList<Ausdruck>();
		liste.add(regel);

		return liste;
	}

	/**
	 * Die Eingangsvariablen der Regel in der Pr&auml;misse zur&uuml;ck.
	 *
	 * @return Menge von Variablennamen
	 */
	public Set<String> getPraemissen() {
		return Ausdruck.Info.getVariablen(praemisse);
	}

	/**
	 * Gibt den Term der Konklusion zur&uuml;ck oder {@code null}, wenn die
	 * Regel nicht anwendbar ist, weil Variablen der Pr&auml;misse undefiniert
	 * sind.
	 *
	 * {@inheritDoc}
	 */
	@Override
	public TermLiteral interpret(final Kontext kontext) {
		assert kontext != null : "Der Kontext darf nicht null sein."; //$NON-NLS-1$

		LogischerWert lw;

		// Variablen der Prämisse prüfen
		for (final String s : getPraemissen()) {
			assert kontext.enthaelt(
					s) : "Alle Variablen einer Regel müssen im Kontext sein. Variable " //$NON-NLS-1$
							+ s + " nicht gefunden."; //$NON-NLS-1$

			if (!(kontext.get(s) instanceof FuzzyVariable)) {
				// Regel ist nicht anwendbar, da Eingangsvariable undefiniert
				return null;
			}
		}

		// Zugehörigkeit des Konklusionsterms bestimmen
		lw = (LogischerWert) praemisse.interpret(kontext);

		// Sicherheitsgrad einbeziehen
		final Ausdruck a = new OperationsSymbol(opImplikation,
				new LogischesLiteral(lw),
				new LogischesLiteral(sicherheitsgrad));
		lw = (LogischerWert) a.interpret(null);

		// Term mit berechnter Zugehörigkeit ausliefern
		return new TermLiteral(konklusion.getName(), konklusion.getTerm(),
				lw.getZugehoerigkeit());
	}

	/**
	 * Gibt die Regel als String zur&uuml;ck.
	 *
	 * {@inheritDoc}
	 *
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		return praemisse + " -> " + konklusion + " (" + sicherheitsgrad + ")"; //$NON-NLS-1$//$NON-NLS-2$//$NON-NLS-3$
	}

	/**
	 * Gibt den eingestellten Implikationsoperator zur&uuml;ck.
	 *
	 * @return Operator
	 */
	Operator getOpImplikation() {
		return opImplikation;
	}

	/**
	 * Legt den f&uuml;r die Implikation zu verwendenden Operator fest. Sinnvoll
	 * ist das Minimum und das algebraische Produkt, aber auch alle anderen
	 * Fuzzy-logischen Operatoren sind erlaubt.
	 *
	 * @param opImplikation
	 *            Operator
	 * @throws NullPointerException
	 *             Wenn der Operator {@code null} ist
	 */
	void setOpImplikation(final Operator opImplikation) {
		assert opImplikation != null;

		this.opImplikation = opImplikation;
	}

}
