/*
 * Segment 5 Intelligente Analyseverfahren, SWE 5.4 Funktionen Fuzzy
 * Copyright (C) 2007-2015 BitCtrl Systems GmbH
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.iav.fuzzylib.fuzzylib.inferenz;

import static de.bsvrz.iav.fuzzylib.interpreter.logik.FuzzyLogikHandler.MAXIMUM;
import static de.bsvrz.iav.fuzzylib.interpreter.logik.FuzzyLogikHandler.MINIMUM;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import de.bsvrz.iav.fuzzylib.fuzzylib.FuzzyVariable;
import de.bsvrz.sys.funclib.bitctrl.interpreter.Ausdruck;
import de.bsvrz.sys.funclib.bitctrl.interpreter.Kontext;
import de.bsvrz.sys.funclib.bitctrl.interpreter.NichtWert;
import de.bsvrz.sys.funclib.bitctrl.interpreter.OperationsSymbol;
import de.bsvrz.sys.funclib.bitctrl.interpreter.Operator;
import de.bsvrz.sys.funclib.bitctrl.interpreter.logik.LogischerWert;
import de.bsvrz.sys.funclib.bitctrl.interpreter.logik.LogischesLiteral;

/**
 * Repr&auml;sentiert eine Fuzzy-Wissensbasis.
 *
 * @author BitCtrl Systems GmbH, Falko Schumann
 */
public class Wissensbasis implements Ausdruck {

	/** Die Liste der Regeln dieser Wissensbasis. */
	private final List<Regel> regeln = new ArrayList<Regel>();

	/** Operator f&uuml;r die Implikation (Standard: Minimum). */
	private Operator opImplikation = MINIMUM;

	/** Operator f&uuml;r die Akkumulation (Standard: Maximum). */
	private Operator opAkkumulation = MAXIMUM;

	/**
	 * F&uuml;gt der Wissensbasis eine Regel hinzu.
	 *
	 * @param regel
	 *            Regel
	 * @return <code>false</code>, wenn die Regel bereits existiert und sie
	 *         nicht hinzugef&uuml;gt wurde
	 */
	public boolean addRegel(final Regel regel) {
		assert regel != null : "Eine Regel darf nicht null sein."; //$NON-NLS-1$

		return regeln.add(regel);
	}

	/**
	 * Gibt die Menge der ableitbaren Fuzzy-Variablen zur&uuml;ck.
	 *
	 * @return Menge mit Variablennamen
	 */
	public Set<String> getKonklusionen() {
		Set<String> liste;

		liste = new HashSet<String>();
		for (final Regel r : regeln) {
			liste.add(r.getKonklusion());
		}

		return liste;
	}

	/**
	 * Gibt die Liste der Regeln zur&uuml;ck.
	 * <p>
	 * {@inheritDoc}
	 */
	@Override
	public List<Ausdruck> getNachfolger() {
		return new ArrayList<Ausdruck>(regeln);
	}

	/**
	 * Gibt den eingestellten Akkumulationsoperator zur&uuml;ck.
	 *
	 * @return Operator
	 */
	public Operator getOpAkkumulation() {
		return opAkkumulation;
	}

	/**
	 * Gibt den eingestellten Implikationsoperator zur&uuml;ck.
	 *
	 * @return Operator
	 */
	public Operator getOpImplikation() {
		return opImplikation;
	}

	/**
	 * Gibt die Menge der Eingangsbvariablen der Wissensbasis zur&uuml;ck.
	 *
	 * @return Menge mit Variablennamen
	 */
	public Set<String> getPraemissen() {
		Set<String> liste;

		liste = new HashSet<String>();
		for (final Regel r : regeln) {
			liste.addAll(r.getPraemissen());
		}

		return liste;
	}

	/**
	 * Gibt die Liste der Regeln in der Wissensbasis zur&uuml;ck.
	 *
	 * @return Liste von Regeln
	 */
	public List<Regel> getRegeln() {
		return new ArrayList<Regel>(regeln);
	}

	/**
	 * Gibt eine Map aller Fuzzy-Variablen zur&uuml;ck, die in den
	 * Regelkonklusionen vorkommen. Ist ein Fuzzy-Wert nicht ableitbar, steht an
	 * der Stelle {@code null} in der Map. Ein Fuzzy-Wert ist nicht ableitbar,
	 * wenn es mindestens eine Regel gibt, die einen Term des Fuzzy-Werts als
	 * Konklusion besitzt, aber nicht anwendbar ist.
	 * <p>
	 * Die Fuzzy-Variablen der Folgerungen werden aus dem Kontext entnommen und
	 * auch dort aktualisiert. Ist eine Fuzzy-Variable nicht ableitbar, ist das
	 * Ergebnis im Kontext undefiniert.
	 *
	 * {@inheritDoc}
	 */
	@Override
	public Map<String, FuzzyVariable> interpret(final Kontext kontext) {
		assert kontext != null : "Der Kontext darf nicht null sein."; //$NON-NLS-1$

		Map<String, FuzzyVariable> folgerungen;

		// Alle Regeln anwenden
		folgerungen = new HashMap<String, FuzzyVariable>();
		for (final Regel regel : regeln) {
			TermLiteral konklusion;

			// Inferenz
			regel.setOpImplikation(opImplikation);
			konklusion = regel.interpret(kontext);
			if (konklusion == null) {
				// Regel ist nicht anwendbar
				folgerungen.put(regel.getKonklusion(), null);
				kontext.set(regel.getKonklusion(), NichtWert.NICHTWERT);
				continue;
			}

			// Akkumulation
			if (!folgerungen.containsKey(konklusion.getName())) {
				// Neue Schlussfolgerung
				FuzzyVariable fv;

				if (!kontext.enthaelt(konklusion.getName())) {
					continue;
				}
				fv = (FuzzyVariable) kontext.get(konklusion.getName());
				fv.setTerm(konklusion.getTerm(),
						konklusion.getZugehoerigkeit());
				folgerungen.put(konklusion.getName(), fv);
			} else if (folgerungen.get(konklusion.getName()) != null) {
				// Akkumulation notwendig
				FuzzyVariable fv;
				Float bisher, neu;
				LogischerWert lw;

				// akkumulierten Wert bestimmen ...
				fv = folgerungen.get(konklusion.getName());
				bisher = fv.getTerm(konklusion.getTerm()).getZugehoerigkeit();
				neu = konklusion.getZugehoerigkeit();
				final Ausdruck a = new OperationsSymbol(opAkkumulation,
						new LogischesLiteral(bisher),
						new LogischesLiteral(neu));
				lw = (LogischerWert) a.interpret(null);

				// ... und setzen
				fv.setTerm(konklusion.getTerm(), lw.getZugehoerigkeit());
			}
		}

		return folgerungen;
	}

	/**
	 * Legt den f&uuml;r die Akkumulation zu verwendenden Operator fest.
	 * Sinnvoll ist das Maximum und die algebraische Summe, aber auch alle
	 * anderen Fuzzy-logischen Operatoren sind erlaubt.
	 *
	 * @param opAkkumulation
	 *            Operator
	 * @throws NullPointerException
	 *             Wenn der Operator null ist
	 */
	public void setOpAkkumulation(final Operator opAkkumulation) {
		assert opAkkumulation != null : "Der Akkumulationsoperator darf nicht null sein."; //$NON-NLS-1$

		this.opAkkumulation = opAkkumulation;
	}

	/**
	 * Legt den f&uuml;r die Implikation zu verwendenden Operator fest. Sinnvoll
	 * ist das Minimum und das algebraische Produkt, aber auch alle anderen
	 * Fuzzy-logischen Operatoren sind erlaubt.
	 *
	 * @param opImplikation
	 *            Operator
	 * @throws NullPointerException
	 *             Wenn der Operator null ist
	 */
	public void setOpImplikation(final Operator opImplikation) {
		assert opImplikation != null : "Der Implikationsoperator darf nicht null sein."; //$NON-NLS-1$

		this.opImplikation = opImplikation;
	}

	/**
	 * Gibt die Wissensbasis als String zur&uuml;ck. Jede Regel ist eine eigene
	 * Zeile.
	 *
	 * {@inheritDoc}
	 *
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		String wb = ""; //$NON-NLS-1$

		for (final Regel r : regeln) {
			wb += r + "\n"; //$NON-NLS-1$
		}

		wb = wb.trim();
		return wb;
	}

}
