/*
 * Segment 5 Intelligente Analyseverfahren, SWE 5.4 Funktionen Fuzzy
 * Copyright (C) 2007-2015 BitCtrl Systems GmbH
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.iav.fuzzylib.interpreter.logik;

import java.util.List;

import de.bsvrz.sys.funclib.bitctrl.interpreter.HandlerValidation;
import de.bsvrz.sys.funclib.bitctrl.interpreter.InterpreterException;
import de.bsvrz.sys.funclib.bitctrl.interpreter.InterpreterMessages;
import de.bsvrz.sys.funclib.bitctrl.interpreter.Operator;
import de.bsvrz.sys.funclib.bitctrl.interpreter.logik.LogikHandler;
import de.bsvrz.sys.funclib.bitctrl.interpreter.logik.LogischerWert;

/**
 * Handler f&uuml;r (fuzzy-)logische Ausdr&uuml;cke.
 *
 * @author BitCtrl Systems GmbH, Falko Schumann
 */
public class FuzzyLogikHandler extends LogikHandler {

	/** Vereinigung bzw. Maximum */
	public static final Operator MAXIMUM = Operator.getOperator("Maximum"); //$NON-NLS-1$

	/** Durchschnitt bzw. Minimum */
	public static final Operator MINIMUM = Operator.getOperator("Minimum"); //$NON-NLS-1$

	/** Komplement&auml;rmenge. */
	public static final Operator KOMPLEMENT = Operator
			.getOperator("Komplement"); //$NON-NLS-1$

	/** Algebraische Summe. */
	public static final Operator ALGEBRAISCHE_SUMME = Operator
			.getOperator("Algebraische Summe"); //$NON-NLS-1$

	/** Algebraisches Produkt. */
	public static final Operator ALGEBRAISCHES_PRODUKT = Operator
			.getOperator("Algebraisches Produkt"); //$NON-NLS-1$

	/** Gamma-Operator. */
	public static final Operator GAMMA_OPERATOR = Operator
			.getOperator("Gamma-Operator"); //$NON-NLS-1$

	/**
	 * Liste der vom Handler unterst&uuml;tzten Operatoren.
	 */
	private static Operator[] operatoren = { UND, ODER, NICHT, MINIMUM, MAXIMUM,
			KOMPLEMENT, IMPLIKATION, ALGEBRAISCHES_PRODUKT, ALGEBRAISCHE_SUMME,
			GAMMA_OPERATOR };

	/**
	 * {@inheritDoc}
	 */
	@Override
	public Operator[] getHandledOperators() {
		return operatoren;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public Object perform(final Operator operator,
			final List<Object> operanden) {
		if ((operator == null)
				|| !validiereHandler(operator, operanden).isValid()) {
			throw new InterpreterException(
					InterpreterMessages.HandlerNotFound.toString());
		}

		Object result = null;

		// Wenn einer der Operanden null ist, ist das Ergebnis auch null
		for (final Object obj : operanden) {
			if (obj == null) {
				return null;
			}
		}

		if ((operator == UND) || (operator == MINIMUM)) {
			result = minimum(operanden.toArray());
		} else if ((operator == ODER) || (operator == MAXIMUM)) {
			result = maximum(operanden.toArray());
		} else if ((operator == NICHT) || (operator == KOMPLEMENT)) {
			result = komplement(operanden.get(0));
		} else if (operator == IMPLIKATION) {
			result = implikation(operanden.toArray());
		} else if (operator == ALGEBRAISCHES_PRODUKT) {
			result = algebraischesProdukt(operanden.toArray());
		} else if (operator == ALGEBRAISCHE_SUMME) {
			result = algebraischeSumme(operanden.toArray());
		} else if (operator == GAMMA_OPERATOR) {
			result = gammaOperator(operanden.toArray());
		}

		return result;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public HandlerValidation validiereHandler(final Operator operator,
			final List<? extends Object> operanden) {
		assert operanden != null : "Liste der Operanden darf nicht null sein."; //$NON-NLS-1$

		boolean anzahlOk = false;
		boolean typOk = true;

		// Anzahl der Operanden und Operation prüfen
		switch (operanden.size()) {
		case 1:
			if (NICHT.equals(operator) || KOMPLEMENT.equals(operator)) {
				anzahlOk = true;
			}
			break;
		case 2:
			if (IMPLIKATION.equals(operator)
					|| ALGEBRAISCHE_SUMME.equals(operator)
					|| ALGEBRAISCHES_PRODUKT.equals(operator)) {
				anzahlOk = true;
			}
			break;
		case 3:
			if (GAMMA_OPERATOR.equals(operator)) {
				anzahlOk = true;
			}
			break;
		default:
			// nix
		}

		if (UND.equals(operator) || ODER.equals(operator)
				|| MAXIMUM.equals(operator) || MINIMUM.equals(operator)) {
			if (operanden.size() >= 1) {
				anzahlOk = true;
			}
		}

		// Alle Operanden müssen ein logischer Wert oder der Nichtwert sein
		for (final Object obj : operanden) {
			if (!(obj instanceof LogischerWert)) {
				typOk = false;
				break;
			}
		}

		return new HandlerValidation(anzahlOk, typOk);
	}

	/**
	 * Berechnet das algebraische Produkt: a * b.
	 *
	 * @param operanden
	 *            Operandenliste mit genau zwei Operanden: a und b
	 * @return Logischer Wert mit berechneter Zugeh&ouml;rigkeit
	 */
	protected static LogischerWert algebraischesProdukt(
			final Object[] operanden) {
		assert operanden != null : "Argument operanden darf nicht null sein."; //$NON-NLS-1$
		assert operanden.length == 2 : "Anzahl der Operanden muss gleich 2 sein."; //$NON-NLS-1$
		assert operanden[0] instanceof LogischerWert : "Operanden müssen logische Werte sein."; //$NON-NLS-1$
		assert operanden[1] instanceof LogischerWert : "Operanden müssen logische Werte sein."; //$NON-NLS-1$

		final Float a = ((LogischerWert) operanden[0]).getZugehoerigkeit();
		final Float b = ((LogischerWert) operanden[1]).getZugehoerigkeit();

		if ((a == null) || (b == null)) {
			return new LogischerWert(null);
		}

		return new LogischerWert(a * b);
	}

	/**
	 * Berechnet die algebraische Summe: a + b - a * b.
	 *
	 * @param operanden
	 *            Operandenliste mit genau zwei Operanden: a und b
	 * @return Logischer Wert mit berechneter Zugeh&ouml;rigkeit
	 */
	protected static LogischerWert algebraischeSumme(final Object[] operanden) {
		assert operanden != null : "Argument operanden darf nicht null sein."; //$NON-NLS-1$
		assert operanden.length == 2 : "Anzahl der Operanden muss gleich 2 sein."; //$NON-NLS-1$
		assert operanden[0] instanceof LogischerWert : "Operanden müssen logische Werte sein."; //$NON-NLS-1$
		assert operanden[1] instanceof LogischerWert : "Operanden müssen logische Werte sein."; //$NON-NLS-1$

		final Float a = ((LogischerWert) operanden[0]).getZugehoerigkeit();
		final Float b = ((LogischerWert) operanden[1]).getZugehoerigkeit();

		if ((a == null) || (b == null)) {
			return new LogischerWert(null);
		}

		return new LogischerWert((a + b) - (a * b));
	}

	/**
	 * F&uuml;hrt die Berechnung mittels Gamma-Operator durch: (a *
	 * b)^(1-&gamma;) * (a + b - a * b)^&gamma;.
	 *
	 * @param operanden
	 *            Operandenliste mit genau drei Operanden: a, b und &gamma;
	 * @return Logischer Wert mit berechneter Zugeh&ouml;rigkeit
	 */
	protected static LogischerWert gammaOperator(final Object[] operanden) {
		assert operanden != null : "Argument operanden darf nicht null sein."; //$NON-NLS-1$
		assert operanden.length == 3 : "Anzahl der Operanden muss gleich 3 sein."; //$NON-NLS-1$
		assert operanden[0] instanceof LogischerWert : "Operanden müssen logische Werte sein."; //$NON-NLS-1$
		assert operanden[1] instanceof LogischerWert : "Operanden müssen logische Werte sein."; //$NON-NLS-1$
		assert operanden[2] instanceof LogischerWert : "Operanden müssen logische Werte sein."; //$NON-NLS-1$

		final Float gamma = ((LogischerWert) operanden[0]).getZugehoerigkeit();
		final Float a = ((LogischerWert) operanden[1]).getZugehoerigkeit();
		final Float b = ((LogischerWert) operanden[2]).getZugehoerigkeit();

		if ((a == null) || (b == null) || (gamma == null)) {
			return new LogischerWert(null);
		}

		final Double d = Math.pow(a * b, 1 - gamma)
				* Math.pow((a + b) - (a * b), gamma);

		return new LogischerWert(d.floatValue());
	}

}
